
# coding: utf-8

# In[1]:

import numpy as np
import matplotlib.pyplot as plt
import time
import math


# In[2]:

def sigmoid(Z):
    return np.divide(1.0, 1.0 + np.exp(-Z))
def sigmoid_derivative(Z):
    S = sigmoid(Z)
    return np.multiply(S, 1.0 - S)
def tanh_derivative(Z):
    return (1.0 - np.tanh(Z)**2)
def ReLU(Z):
    return np.maximum(0, Z)
def ReLU_derivative(Z):
    return np.int64(Z > 0)
def activation(Z, activation_function_name = "sigmoid"):
    if activation_function_name == "sigmoid":
        S = sigmoid(Z)
    elif activation_function_name == "tanh":
        S = np.tanh(Z)
    elif activation_function_name == "ReLU":
        S = ReLU(Z)
    return S
def activation_derivative(Z, activation_function_name = "sigmoid"):
    if activation_function_name == "sigmoid":
        dS = sigmoid_derivative(Z)
    elif activation_function_name == "tanh":
        dS = tanh_derivative(Z)
    elif activation_function_name == "ReLU":
        dS = ReLU_derivative(Z)
    return dS
def normalize_X_Y(X, Y, X_test, Y_test):
    Xu = np.mean(X, axis = 1, keepdims = True)
    Xs = np.std(X, axis = 1, keepdims = True)
    XX = (X - Xu) / Xs
    YY = np.int64(Y > 0.5)
    XX_test = (X_test - Xu) / Xs
    if len(Y_test) != 0:
        YY_test = np.int64(Y_test > 0.5)
    else:
        YY_test = []
    return XX, Xu, Xs, YY, XX_test, YY_test
def normalize_W_b(W, b, Xu, Xs): # modified 12/30/2017
    L = len(W)
    WW = [[]] * L
    bb = [[]] * L
    WW[1] = W[1] * Xs.T
    bb[1] = b[1] + np.dot(W[1], Xu)
    for l in range(2, L):
        WW[l] = np.copy(W[l])
        bb[l] = np.copy(b[l])
    return WW, bb
def denormalize_WW_bb(WW, bb, Xu, Xs): # modified 12/30/2017
    L = len(WW)
    W = [[]] * L
    b = [[]] * L
    W[1] = WW[1] / Xs.T
    b[1] = bb[1] - np.dot(WW[1], Xu / Xs)
    for l in range(2, L):
        W[l] = np.copy(WW[l])
        b[l] = np.copy(bb[l])
    return W, b
def initialize_WW_bb(NN, hlayer_activation="ReLU"):
    L = len(NN)
    WW = [[]] * L   # 1st element is empty
    bb = [[]] * L   # 1st element is empty
    if hlayer_activation == "ReLU":
        COEFF = 2.0 # He initialization for ReLU activations
    else:
        COEFF = 1.0 # tanh activation, etc.
    for l in range(1, L):
        WW[l] = np.random.randn(NN[l], NN[l-1]) * np.sqrt(COEFF / NN[l-1])
        bb[l] = np.zeros((NN[l], 1))
    return WW, bb
def initialize_vv_ss(NN):
    L = len(NN)
    vvdW = [[]] * L   # Adam optimization
    vvdb = [[]] * L   # Adam optimization
    ssdW = [[]] * L   # Adam optimization
    ssdb = [[]] * L   # Adam optimization
    for l in range(1, L):
        vvdW[l] = np.zeros((NN[l], NN[l-1]))
        vvdb[l] = np.zeros((NN[l], 1))
        ssdW[l] = np.zeros((NN[l], NN[l-1]))
        ssdb[l] = np.zeros((NN[l], 1))
    return vvdW, vvdb, ssdW, ssdb
def forward_propagation(X, W, b, hlayer_activation, keep_prob):
    L = len(W)
    Z = [[]] * L
    A = [[]] * L
    DO = [[]] * L # drop-out regularization
    A[0] = X
    if keep_prob[0] >= 1.0:  # drop-out regularization
        DO[0] = 1
    else:
        DO[0] = np.random.rand(A[0].shape[0], A[0].shape[1]) < keep_prob[0]
        DO[0] = np.divide(DO[0] , keep_prob[0])
        A[0] = np.multiply(A[0], DO[0])
    for l in range(1, L):
        Z[l] = np.dot(W[l], A[l-1]) + b[l]
        if l < (L-1):
            A[l] = activation(Z[l], hlayer_activation)
        else:
            A[l] = activation(Z[l], "sigmoid")
        if keep_prob[l] >= 1.0:  # drop-out regularization
            DO[l] = 1
        else:
            DO[l] = np.random.rand(A[l].shape[0], A[l].shape[1]) < keep_prob[l]
            DO[l] = np.divide(DO[l] , keep_prob[l])
            A[l] = np.multiply(A[l], DO[l])
    return Z, A, DO
def backward_propagation(Y, W, A, Z, DO, hlayer_activation, L2_regularization_beta):
    m = Y.shape[1]
    L = len(W)
    dZ = [[]] * L
    dA = [[]] * L
    dW = [[]] * L
    db = [[]] * L
    dZ[-1] = (1.0/m) * (A[-1] - Y) # Sigmoid function is assumed for the last layer activation
    dZ[-1] = np.multiply(dZ[-1], DO[-1]) # drop-out regularization
    for l in reversed(range(1, L)):
        dW[l] = np.dot(dZ[l], A[l-1].T)
        if L2_regularization_beta != 0:
            dW[l] = dW[l] + (L2_regularization_beta / m) * W[l]  # L2 regularization added
        db[l] = np.sum(dZ[l], axis=1, keepdims=True)
        if l != 1:
            dA[l-1] = np.dot(W[l].T, dZ[l])
            dA[l-1] = np.multiply(dA[l-1], DO[l-1]) # drop-out regularization
            dZ[l-1] = np.multiply(dA[l-1], activation_derivative(Z[l-1], hlayer_activation))
    return dW, db
def compute_cost(Y, AL, WW, L2_regularization_beta):
    m = Y.shape[1]
    ii = (Y > 0.5)
    LOSS = np.zeros(Y.shape)
    LOSS[ii] = - np.log(AL[ii])   # Y == 1
    LOSS[~ii] = - np.log(1 - AL[~ii])   # Y == 0
    cost_l2_regularization = 0.0
    if L2_regularization_beta != 0:
        for l in range(1,len(WW)):
            cost_l2_regularization = cost_l2_regularization + (L2_regularization_beta / 2 / m) * np.sum(np.square(WW[l]))
    cost = (1/m) * np.sum(LOSS) + cost_l2_regularization
    cost = cost.squeeze()
    assert(isinstance(cost,float) or isinstance(cost,int))
    return cost
def update_WW_bb(opt_method, W, b, dW, db, vvdW, vvdb, ssdW, ssdb, learning_rate, adam_cnt, adam_beta1, adam_beta2, adam_epsilon):
    L = len(W)
    vdW_corrected = [[]] * L # Adam optimization
    vdb_corrected = [[]] * L # Adam optimization
    sdW_corrected = [[]] * L # Adam optimization
    sdb_corrected = [[]] * L # Adam optimization
    for l in range(1, L):
        if (opt_method == "momentum") or (opt_method == "adam"):
            vvdW[l] = adam_beta1 * vvdW[l] + (1 - adam_beta1) * dW[l] # Adam optimization
            vvdb[l] = adam_beta1 * vvdb[l] + (1 - adam_beta1) * db[l] # Adam optimization
            vdW_corrected[l] = vvdW[l] / (1 - adam_beta1**adam_cnt) # Adam optimization
            vdb_corrected[l] = vvdb[l] / (1 - adam_beta1**adam_cnt) # Adam optimization
        if opt_method == "adam":
            ssdW[l] = adam_beta2 * ssdW[l] + (1 - adam_beta2) * (dW[l])**2 # Adam optimization
            ssdb[l] = adam_beta2 * ssdb[l] + (1 - adam_beta2) * (db[l])**2 # Adam optimization
            sdW_corrected[l] = ssdW[l] / (1 - adam_beta2**adam_cnt) # Adam optimization
            sdb_corrected[l] = ssdb[l] / (1 - adam_beta2**adam_cnt) # Adam optimization
        #--------------------------------------------
        if opt_method == "gradient decent":
            W[l] = W[l] - learning_rate * dW[l]
            b[l] = b[l] - learning_rate * db[l]
        elif opt_method == "momentum":
            W[l] = W[l] - learning_rate * vdW_corrected[l] 
            b[l] = b[l] - learning_rate * vdb_corrected[l]    
        elif opt_method == "adam":
            W[l] = W[l] - learning_rate * vdW_corrected[l] / (np.sqrt(sdW_corrected[l]) + adam_epsilon)
            b[l] = b[l] - learning_rate * vdb_corrected[l] / (np.sqrt(sdb_corrected[l]) + adam_epsilon)
    return W, b, vvdW, vvdb, ssdW, ssdb
def predict(X, Y, W, b, hlayer_activation):
    if len(X) != 0:
        _, A, _ = forward_propagation(X, W, b, hlayer_activation, [1.0]*len(W)) # [1.0]*leng(W): turn off drop-out regularization
        Y_predict = np.int64(A[-1] > 0.5)
    else:
        Y_predict = []
    if len(Y) != 0:
        predict_accuracy = 1 - np.mean(np.abs(Y - Y_predict))
    else:
        predict_accuracy = np.nan
    return Y_predict, predict_accuracy
def plot_contour(w, b, hlayer_activation):
    # https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.meshgrid.html
    N = 100   # the number of grid points
    x = np.linspace(0,1,N)
    XV, YV = np.meshgrid(x, x)
    xy = np.array([XV.reshape(1,-1).squeeze(), YV.reshape(1,-1).squeeze()])
    zz, aa, _ = forward_propagation(xy, w, b, hlayer_activation, [1.0]*len(w)) # [1.0]*leng(w): turn off drop-out regularization
    ZY = zz[-1].reshape(XV.shape)
    # https://matplotlib.org/examples/pylab_examples/contour_demo.html
    CS = plt.contour(XV, YV, ZY, 6, colors="k", levels=[0])  # draw only zero line
    #plt.clabel(CS, fontsize=9, inline=1)
def DEBUG_gradcheck_Wb_to_vector(W, b): # modified 12/30/2017
    theta = []
    for l in range(1,len(W)):
        theta.extend(W[l].reshape(-1,1))
        theta.extend(b[l].reshape(-1,1))
    theta = np.array(theta)
    assert(theta.shape == (len(theta),1))
    return theta
def DEBUG_gradcheck_vector_to_Wb(theta, W, b):
    WW = [[]]*len(W)
    bb = [[]]*len(b)
    n = 0
    for l in range(1,len(W)):
        LEN1 = W[l].shape[0] * W[l].shape[1]
        WW[l] = theta[n:(n+LEN1)].reshape(W[l].shape)
        n = n + LEN1
        LEN2 = b[l].shape[0] * b[l].shape[1]
        bb[l] = theta[n:(n+LEN2)].reshape(b[l].shape)
        n = n + LEN2
    return WW, bb
def DEBUG_gradcheck(X, Y, W, b, hlayer_activation, L2_regularization_beta):
    epsilon = 1e-7
    theta = DEBUG_gradcheck_Wb_to_vector(W, b)
    gradapprox = np.zeros(theta.shape)
    for i in range(len(theta)):
        thetaplus = np.copy(theta)
        thetaplus[i] = thetaplus[i] + epsilon
        W1, b1 = DEBUG_gradcheck_vector_to_Wb(thetaplus, W, b)
        _, A1, _ = forward_propagation(X, W1, b1, hlayer_activation, [1]*len(W))
        Jplus = compute_cost(Y, A1[-1], W1, L2_regularization_beta)
        #-----------------------------------------------------
        thetaminus = np.copy(theta)
        thetaminus[i] = thetaminus[i] - epsilon
        W2, b2 = DEBUG_gradcheck_vector_to_Wb(thetaminus, W, b)
        _, A2, _ = forward_propagation(X, W2, b2, hlayer_activation, [1]*len(W))
        Jminus = compute_cost(Y, A2[-1], W2, L2_regularization_beta)
        #-----------------------------------------------------
        gradapprox[i] = (Jplus - Jminus) / 2 / epsilon
    Z, A, DO = forward_propagation(X, W, b, hlayer_activation, [1]*len(W))
    dW, db = backward_propagation(Y, W, A, Z, DO, hlayer_activation, L2_regularization_beta)
    grad = DEBUG_gradcheck_Wb_to_vector(dW, db)
    #---------------------------------------------------
    numerator = np.linalg.norm(gradapprox - grad, 2) # L2 norm
    denominator = np.linalg.norm(gradapprox, 2) + np.linalg.norm(grad, 2) # L2 norm
    difference = numerator / denominator
    if difference > 2e-7:
        print ("\033[93m" + ">> DEBUG: There is a mistake in the backward propagation! difference = " + str(difference) + "\033[0m")
    else:
        print ("\033[92m" + ">> DEBUG: Your backward propagation works perfectly fine! difference = " + str(difference) + "\033[0m")
    return difference
def random_mini_batches(X, Y, mini_batch_size = 64):
    #np.random.seed(seed)            # To make your "random" minibatches the same as ours
    m = X.shape[1]                  # number of training examples
    mini_batches = []
    # Step 1: Shuffle (X, Y)
    permutation = np.random.permutation(m)
    shuffled_X = X[:, permutation]
    shuffled_Y = Y[:, permutation]
    # Step 2: Partition (shuffled_X, shuffled_Y). Minus the end case.
    num_complete_minibatches = max(1, math.floor(m/mini_batch_size)) # number of mini batches of size mini_batch_size in your partitionning
    for k in range(0, num_complete_minibatches):
        mini_batch_X = shuffled_X[:,(k*mini_batch_size):((k+1)*mini_batch_size)]
        mini_batch_Y = shuffled_Y[:,(k*mini_batch_size):((k+1)*mini_batch_size)]
        mini_batch = (mini_batch_X, mini_batch_Y)
        mini_batches.append(mini_batch)
    # Handling the end case (last mini-batch < mini_batch_size)
    if (m >= mini_batch_size) and (m % mini_batch_size != 0):
        mini_batch_X = shuffled_X[:,(num_complete_minibatches*mini_batch_size):]
        mini_batch_Y = shuffled_Y[:,(num_complete_minibatches*mini_batch_size):]
        mini_batch = (mini_batch_X, mini_batch_Y)
        mini_batches.append(mini_batch)
    return mini_batches
def model(X, Y, X_test, Y_test, W = [], b = [], hidden_layer_dims = [], num_epochs = 2000, learning_rate = 0.001, hlayer_activation = "ReLU", 
          L2_regularization_beta = 0.0, DropOut_regularization_keep_prob = 1.0, mini_batch_size = 64, 
          opt_method = "adam", adam_beta1 = 0.9, adam_beta2 = 0.999, adam_epsilon = 1e-8):
    print("> Deep Neural Network started...")
    print(">> version = 2017-12-30")
    print(">> the number of training samples = " + str(X.shape[1]))
    print(">> the number of test samples = " + str(X_test.shape[1] if len(X_test)!=0 else X_test))
    NN = [X.shape[0]] + hidden_layer_dims + [Y.shape[0]]
    print(">> the number of units of layers = " + str(NN))
    print(">> the number of epochs = " + str(num_epochs))
    print(">> learning rate = " + str(learning_rate))
    print(">> activation function of hidden layers = " + hlayer_activation)
    print(">> L2 regularization: beta = " + str(L2_regularization_beta))
    if isinstance(DropOut_regularization_keep_prob,float) or isinstance(DropOut_regularization_keep_prob,int):
        keep_prob = [DropOut_regularization_keep_prob] * len(NN)
    elif len(DropOut_regularization_keep_prob) == len(NN):
        keep_prob = DropOut_regularization_keep_prob
    print(">> Drop-Out regularization: keep_prob = " + str(keep_prob))
    print(">> mini batch size = " + str(mini_batch_size))
    print(">> optimization method = " + opt_method)
    print(">> Adam optimization: beta1 = " + str(adam_beta1))
    print(">> Adam optimization: beta2 = " + str(adam_beta2))
    print(">> Adam optimization: epsioin = " + str(adam_epsilon))
    #------------------------------------------------------------------------------
    XX, Xu, Xs, YY, XX_test, YY_test = normalize_X_Y(X, Y, X_test, Y_test)
    if (len(W) == 0) or (len(b) == 0):
        WW, bb = initialize_WW_bb(NN, hlayer_activation)
        print(">> W and b are intialized...")
    else:
        WW, bb = normalize_W_b(W, b, Xu, Xs)
        print(">> User-defined W and b are given and normalized...")
    vvdW, vvdb, ssdW, ssdb = initialize_vv_ss(NN)
    costs = []
    train_accuracies = []
    test_accuracies = []
    adam_cnt = 0;
    tic = time.time()
    for i in range(1, num_epochs+1):
        #seed = seed + 1
        #minibatches = random_mini_batches(X, Y, mini_batch_size, seed)
        minibatches = random_mini_batches(XX, YY, mini_batch_size)
        # minibatches = [(XX,YY)]
        for minibatch in minibatches:
            (minibatch_X, minibatch_Y) = minibatch
            ZZ, AA, DO = forward_propagation(minibatch_X, WW, bb, hlayer_activation, keep_prob)
            cost = compute_cost(minibatch_Y, AA[-1], WW, L2_regularization_beta)
            dW, db = backward_propagation(minibatch_Y, WW, AA, ZZ, DO, hlayer_activation, L2_regularization_beta)
            adam_cnt = adam_cnt + 1
            WW, bb, vvdW, vvdb, ssdW, ssdb = update_WW_bb(opt_method, WW, bb, dW, db, vvdW, vvdb, ssdW, ssdb, learning_rate, adam_cnt, adam_beta1, adam_beta2, adam_epsilon)
        costs.append(cost)
        Yp_train, train_accuracy = predict(XX, YY, WW, bb, hlayer_activation)
        train_accuracies.append(train_accuracy)
        Yp_test, test_accuracy = predict(XX_test, YY_test, WW, bb, hlayer_activation)
        test_accuracies.append(test_accuracy)
        if (i == 1) or (i%1000 == 0) or (i == num_epochs):
            toc = time.time()
            print(">> epoch = {}; adam_count = {}; cost = {:.6f}; train_accuracy = {:.3f}; test_accuracy = {:.3f}; time = {} sec"
                  .format(i, adam_cnt, cost, train_accuracy, test_accuracy, toc-tic))
            tic = time.time()
    #------------------------------------------------------------------------------
    DEBUG_gradcheck(XX, YY, WW, bb, hlayer_activation, L2_regularization_beta) # DEBUGGING
    W, b = denormalize_WW_bb(WW, bb, Xu, Xs)
    fout = {"costs":costs,
            "train_accuracies":train_accuracies,
            "test_accuracies":test_accuracies,
            "W":W,
            "b":b,
            "Yp_train":Yp_train,
            "Yp_test":Yp_test}
    return fout


# In[3]:

if False:
    M = 1000   # the number of samples
    np.random.seed(1)
    X = np.random.rand(2, M)
    Y = np.zeros((1, M))
    #Y[0, (X[1,:] - X[0,:] + 0.3) > 0 ] = 1  # line
    #Y[0, (X[0,:]**2 + X[1,:]**2 - 0.7**2) > 0 ] = 1  # semi-circle
    #Y[0, ((X[0,:]-0.5)**2 + (X[1,:]-0.5)**2 - 0.3**2) > 0] = 1  # circle
    #Y[0, (((X[0,:]-0.5)/0.4)**2 + ((X[1,:]-0.2)/0.05)**2 - 1) > 0] = 1  # ellipse
    Y[0, ((((X[0,:]-0.5)/0.4)**2 + ((X[1,:]-0.2)/0.05)**2 - 1) > 0) & (((X[0,:]-0.7)**2 + (X[1,:]-0.7)**2 - 0.1**2) > 0)] = 1  # ellipse & circle
    #-----------------------------------------------------------------
    rn = np.random.permutation(M)
    NofT = math.floor(M * 0.7) # number of train samples
    X_train = X[:,rn[0:NofT]]
    Y_train = Y[:,rn[0:NofT]]
    X_test = X[:,rn[NofT:]]
    Y_test = Y[:,rn[NofT:]]
    #-----------------------------------------------------------------
    hidden_layer_dims = [10,5,2]
    num_epochs = 1000
    learning_rate = 0.001
    hlayer_activation = ["sigmoid", "tanh", "ReLU"][2]
    L2_regularization_beta = 0.0
    DropOut_regularization_keep_prob = 1.0# [1.0, 0.7, 0.8, 0.9, 1.0]
    mini_batch_size = 64
    opt_method = ["gradient decent", "momentum", "adam"][2]
    adam_beta1 = 0.9
    adam_beta2 = 0.999
    adam_epsilon = 1e-8
    fout1 = model(X_train, Y_train, [], [], [], [], hidden_layer_dims, num_epochs, learning_rate, hlayer_activation, L2_regularization_beta, DropOut_regularization_keep_prob, mini_batch_size, opt_method, adam_beta1, adam_beta2, adam_epsilon)
    fout2 = model(X_train, Y_train, X_test, Y_test, fout1["W"], fout1["b"], hidden_layer_dims, num_epochs, learning_rate, hlayer_activation, L2_regularization_beta, DropOut_regularization_keep_prob, mini_batch_size, opt_method, adam_beta1, adam_beta2, adam_epsilon)
    #-----------------------------------------------------------------
    get_ipython().magic('matplotlib inline')
    plt.scatter(X_train[0,:], X_train[1,:], c=Y_train, s=20, cmap=plt.cm.Spectral)
    plot_contour(fout2["W"], fout2["b"], hlayer_activation)    # user-defined plot
    plt.show()
    plt.scatter(X_test[0,:], X_test[1,:], c=Y_test, s=20, cmap=plt.cm.Spectral)
    plot_contour(fout2["W"], fout2["b"], hlayer_activation)    # user-defined plot
    plt.show()
    plt.plot(fout1["costs"] + fout2["costs"])
    plt.plot(fout1["costs"])
    plt.show()
    plt.plot(fout1["train_accuracies"] + fout2["train_accuracies"],'b-')
    plt.plot(fout1["train_accuracies"],'k-')
    plt.plot(fout1["test_accuracies"] + fout2["test_accuracies"],'r-')
    plt.plot(fout1["test_accuracies"],'k-')
    plt.show()
    


# In[ ]:






#!/usr/bin/env python
# coding: utf-8

# In[1]:


__version__ = '2019-07-27-beta'
__DEBUG_MODE__ = False


# In[2]:


# Copyright © 2018-2019 Bomsoo Brad Kim, All Rights Reserved.
# Last Update : 04/14/2019
import numpy as np
import pandas as pd

def xgrid_search(xgrid_search_target_function, params, 
                 MIN_MAX = 'min', SCORE_TOLERANCE = 1e-5,  MAX_GRID_LEVEL = 20, MAX_INNER_ROUND = 50, 
                 initialize_cross_grid = False):
    def generate_cgrid(ikeys): # make cross grids
        iparams2 = {k: [0,-1,1] for k in ikeys}
        cgrid = pd.DataFrame(columns = ikeys) # initialize cgrid with empty elements
        for k,i in iparams2.items():
            cgrid = pd.concat([cgrid, pd.DataFrame({k:i})], sort = False) # make cross grids
        cgrid.fillna(0, inplace = True)
        cgrid.drop_duplicates(keep = 'first', inplace = True)
        cgrid.reset_index(drop = True, inplace = True)
        return cgrid

    def trim_grid(grid, ikeys, min_grid, max_grid): # remove duplicate and out-of-range grid points
        i0 = grid[ikeys].duplicated(keep = 'first') # find out which are duplicate elements except the first elements
        i1 = (grid[ikeys] < min_grid[ikeys]).any(axis = 1) # find out rows with less than min indexes
        i2 = (grid[ikeys] > max_grid[ikeys]).any(axis = 1) # find out rows with greater than max indexes
        # grid = pd.concat([grid, i0, i1, i2], sort = False, ignore_index = True, axis = 1)
        grid = grid.loc[~(i0 | i1 | i2)] # remove duplicate and out-of-range grid points
        grid.reset_index(drop = True, inplace = True)
        return grid

    def fill_keys(params2, grid):
        for (k,i) in params2.items(): grid[k] = grid['i_'+k].map({i:value for i, value in enumerate(params2[k])}) # update index
        return grid

    def update_divide_intervals(partype, params2, grid):
        def divide_intervals(a, paramter_type = 'uni'):
            a = np.array(a)
            if paramter_type == 'uni': aa = np.sort(np.unique(np.append(a, (a[1:] + a[:-1])/2)))
            elif paramter_type == 'log': aa = np.sort(np.unique(np.append(a, np.sqrt(a[1:]*a[:-1]))))
            elif paramter_type == 'int': aa = np.sort(np.unique(np.around(np.append(a, (a[1:] + a[:-1])/2)).astype(int)))
            return aa
        for (k,i) in params2.items():
            params2[k] = divide_intervals(i, partype[k]) # divide intervals
            grid['i_'+k] = grid[k].map({value:i for i, value in enumerate(params2[k])}) # update index
        return params2, grid
    
    
    # initial preparation
    partype = {k:i[0] for (k,i) in params.items()} # pick up data type
    params2 = {k:np.sort(np.unique(i[1])) for (k,i) in params.items()} # pick up range
    iparams = {'i_'+k:range(len(i)) for (k,i) in params2.items()} # index for parameters
    pkeys = list(params2.keys())
    ikeys = list(iparams.keys()) # parameter key names
    cgrid = generate_cgrid(ikeys) # cross grid
    
    # initialize grid
    grid = pd.DataFrame(index = pd.MultiIndex.from_product(iparams.values(), names = iparams.keys())).reset_index() # https://stackoverflow.com/questions/13269890/cartesian-product-in-pandas
    grid['func_value'] = np.nan
    grid = fill_keys(params2, grid)
    
    # cross grid setting in the first place
    if initialize_cross_grid:
        params2, grid = update_divide_intervals(partype, params2, grid)
        grid = grid[ikeys].mean(axis = 0).astype(int) + cgrid
        grid[ikeys].astype(int)
        grid['func_value'] = np.nan
        grid = fill_keys(params2, grid)

    # search for the best
    best_fval = pd.DataFrame()
    len_prev_grid = 0
    for nn in range(MAX_GRID_LEVEL):
        for kk in range(MAX_INNER_ROUND):
            # evaluate function
            for i in range(len(grid)):
                if np.isnan(grid.loc[i,'func_value']): # if func_value = nan, then evaluate
                    grid.loc[i,'func_value'] = xgrid_search_target_function(**grid.astype(object).loc[i, pkeys].to_dict())
                    print('> i=%s (level=%s, group=%s); '%(i+1,nn+1,kk+1), grid.astype(object).loc[i, pkeys+['func_value']].to_dict())
            # find best score
            if MIN_MAX == 'max': idx = grid['func_value'].idxmax()
            elif MIN_MAX == 'min': idx = grid['func_value'].idxmin()
            best_grid = grid.loc[idx, ikeys].astype('int') # ensure that the index is integer
            print('  best so far =', grid.astype(object).loc[idx, pkeys+['func_value']].to_dict())
            # explore new grid points
            min_grid = grid.min(axis = 0) # order matters!
            max_grid = grid.max(axis = 0) # order matters!
            grid = pd.concat([grid, best_grid + cgrid], sort = False, ignore_index = True) # attach new rows
            grid = trim_grid(grid, ikeys, min_grid, max_grid) # remove duplicate and out-of-range rows
            grid = fill_keys(params2, grid)
            # stop if there is no NAN value
            if ~grid['func_value'].isna().any(): break

        # stop condition
        if len_prev_grid == len(grid):
            break
        else:
            if MIN_MAX == 'max': 
                idx = grid.loc[len_prev_grid:, 'func_value'].idxmax() # find the best at the latest level
                best_fval = best_fval.append(grid.loc[idx, pkeys+['func_value']]) # save the best row
                best_fval.sort_values(by = 'func_value', ascending  = False, inplace = True)
            elif MIN_MAX == 'min': 
                idx = grid.loc[len_prev_grid:, 'func_value'].idxmin() # find the best at the latest level
                best_fval = best_fval.append(grid.loc[idx, pkeys+['func_value']]) # save the best row
                best_fval.sort_values(by = 'func_value', ascending  = True, inplace = True)
            best_fval.drop_duplicates(keep = 'first', inplace = True)
            if len(best_fval) >= 2:
                best_fval_values = best_fval['func_value'].values
                if abs((best_fval_values[0] - best_fval_values[1])/best_fval_values[0]) < SCORE_TOLERANCE:
                    print('cross grid search completed...'); break;
        len_prev_grid = len(grid)

        # devide intervals
        if nn < MAX_GRID_LEVEL - 1: params2, grid = update_divide_intervals(partype, params2, grid)
                
    if MIN_MAX == 'max': idx = grid['func_value'].idxmax()
    elif MIN_MAX == 'min': idx = grid['func_value'].idxmin()
    best_param = grid.astype(object).loc[idx, pkeys].to_dict()
    
    grid.drop(columns = ikeys, inplace = True)
    return best_param, best_fval, grid

def left_join_crossgridparams_params(xgrid_params, params):
    for (k,i) in xgrid_params.items(): 
        xgrid_params[k][1] = np.sort(np.unique(np.append(xgrid_params[k][1], params[k])))
    return xgrid_params

#### Getting Started! ##############################################
if __DEBUG_MODE__:
    def xgrid_search_target_function(**param):
        #--- STRAT user definition ---
        def any_user_function(x,y,z):
            return ((x-2.7)**2) + ((y-3.3)**2) + ((z+2.6)**2) + 1.0
        output = any_user_function(**param)
        #--- END user definition ---
        return output

    params = { # define the data type and input varaible range
        'x': ['int', [-10, 10]], # uni / log / int
        'y': ['log', [0, 0.01, 10]], # uni / log / int
        'z': ['uni', [-10, -1, 10]]  # uni / log / int
    }

    param = {'x': 3, 'y': 4, 'z': 5, 'a': 6, 'b': 7} # sample code for how to use left_join_crossgridparams_params
    params = left_join_crossgridparams_params(params, param) # sample code for how to use left_join_crossgridparams_params

    best_param, best_fval, grid = xgrid_search(xgrid_search_target_function, params, MIN_MAX = 'min', SCORE_TOLERANCE = 1e-5) # decide on min/max problem and then run!
    print('best param = ',best_param)
    print(best_fval)


# In[3]:


import pandas as pd
import numpy as np
import time
import xgboost as xgb
from xgboost.sklearn import XGBClassifier, XGBRegressor
import lightgbm as lgb
from lightgbm.sklearn import LGBMClassifier, LGBMRegressor
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.model_selection import ParameterGrid
from sklearn.model_selection import train_test_split
from sklearn import metrics   #Additional scklearn functions

def xgrid_search_boosting(train, test, features, target, params_schedule, model, eval_metric, MIN_MAX, 
                          CV_MODE = 'cross validation 1', N_FOLD = 5, N_BOOST_ROUND = 10000, EARLY_STOPPING = 50, RAND_SEED = 123): 
    #--- user_defined_eval_function --------------------------------------------------------------
    def user_defined_eval_function(train, test, features, target, model, eval_metric, MIN_MAX, predict_test_output = False):
        if CV_MODE == 'validation': # validatoin approach: XGBoost
            X_train, X_valid, y_train, y_valid = train_test_split(train[features], train[target], test_size=1/N_FOLD, random_state=RAND_SEED)
            model.set_params(n_estimators = N_BOOST_ROUND) # initialize n_estimators
            model.fit(X_train, y_train, eval_set = [(X_train, y_train), (X_valid, y_valid)], eval_metric = eval_metric, 
                        early_stopping_rounds = EARLY_STOPPING, verbose = False) #Fit the algorithm on the data
            if MIN_MAX == 'max': idx = np.array(list(model.evals_result()['validation_1'].values())).argmax()
            elif MIN_MAX == 'min': idx = np.array(list(model.evals_result()['validation_1'].values())).argmin()
            model.set_params(n_estimators = idx + 1) # update n_estimators
            train_score = np.array(list(model.evals_result()['validation_0'].values())).squeeze()[idx]
            valid_score = np.array(list(model.evals_result()['validation_1'].values())).squeeze()[idx]
            test_pred = model.predict(test[features])     # computationally not expensive
        elif (CV_MODE == 'cross validation 1') and ('XGB' in str(model)): # cross validation 1: XGBoost
            dtrain = xgb.DMatrix(train[features], label=train[target], missing=np.nan) # missing value handling: https://www.youtube.com/watch?v=cVqDguNWh4M
            cvoutp = xgb.cv(model.get_xgb_params(), dtrain, num_boost_round=N_BOOST_ROUND, verbose_eval=False,
                              nfold=N_FOLD, metrics=eval_metric, early_stopping_rounds=EARLY_STOPPING, seed=RAND_SEED) 
            model.set_params(n_estimators = cvoutp.shape[0]) # update n_estimator 
            train_score = cvoutp.tail(1)[cvoutp.columns[cvoutp.columns.str.contains('train-.+-mean', regex=True)]].squeeze()
            valid_score = cvoutp.tail(1)[cvoutp.columns[cvoutp.columns.str.contains('test-.+-mean', regex=True)]].squeeze()

            if predict_test_output == True:
                model.fit(train[features], train[target].values.ravel(), eval_metric = eval_metric) #Fit the algorithm on the data
                test_pred = model.predict(test[features])    
            else: 
                test_pred = []
        elif (CV_MODE == 'cross validation 1') and ('LGB' in str(model)): # cross validation 1: LightGBM
            dtrain = lgb.Dataset(train[features], label=train[target])

            cvoutp = lgb.cv({k:v for k,v in model.get_params().items() if k not in ['n_estimators', 'silent']}, # exclude n_estimators because of the argument, 'num_boost_round'
                            dtrain, num_boost_round=N_BOOST_ROUND, verbose_eval=False, 
                            nfold=N_FOLD, metrics=eval_metric, early_stopping_rounds=EARLY_STOPPING, seed=RAND_SEED)
            model.set_params(n_estimators = len(cvoutp[eval_metric+'-mean'])) # update n_estimator with the best num_boost_round
            valid_score = cvoutp[eval_metric+'-mean'][-1] # best CV score

            if predict_test_output == True:
                model.fit(train[features], train[target].values.ravel(), eval_metric = eval_metric) #Fit the algorithm on the data
                test_pred = model.predict(test[features])    
            else: 
                test_pred = []
        elif CV_MODE == 'cross validation 2': # cross validation 2: XGBoost, LightGBM
#             folds = StratifiedKFold(n_splits=N_FOLD, shuffle=False, random_state=RAND_SEED) # cv n-fold
            folds = KFold(n_splits=N_FOLD, shuffle=False, random_state=RAND_SEED) # cv n-fold
            oof = np.zeros(len(train))
            test_pred = np.zeros(len(test))
            for n, (trn_idx, val_idx) in enumerate(folds.split(train[features].values, train[target].values)):
                X_train, y_train = train.iloc[trn_idx][features], train.iloc[trn_idx][target].values.ravel()
                X_valid, y_valid = train.iloc[val_idx][features], train.iloc[val_idx][target].values.ravel()

                model.set_params(n_estimators = N_BOOST_ROUND) # initialize n_estimators
                model.fit(X_train, y_train, eval_set = [(X_train, y_train), (X_valid, y_valid)], eval_metric = eval_metric, 
                            early_stopping_rounds = EARLY_STOPPING, verbose = False) #Fit the algorithm on the data

                if eval_metric == 'auc': 
                    oof[val_idx] = model.predict_proba(X_valid)[:,1]
                    test_pred += model.predict_proba(test[features])[:,1] / folds.n_splits
                if eval_metric == 'rmse': 
                    oof[val_idx] = model.predict(X_valid)
                    test_pred += model.predict(test[features]) / folds.n_splits

            if eval_metric == 'auc': 
                valid_score = metrics.roc_auc_score(train[target], oof)
            if eval_metric == 'rmse': 
                valid_score = np.sqrt(metrics.mean_squared_error(train[target], oof))
        return test_pred, valid_score    
    
    #--- evaluation function with xgboost, ligthgbm ------------------------------
    def xgrid_search_target_function(**param): # define user function with all the input variables
        #--- START: user defined function -----------------------------
        model.set_params(**param) # update some parameters
        test_pred, valid_score = user_defined_eval_function(train, test, features, target, model, eval_metric, MIN_MAX)
        #--- END ------------------------------------------------------
        return valid_score

    #-----------------------------------------------------------------------
    tic = time.time()
    for params in params_schedule:
    #     params = left_join_crossgridparams_params(params, model.get_xgb_params()) # ensure that the latest xgmodel values are included
        best_param, best_fval, grid = xgrid_search(xgrid_search_target_function, params, MIN_MAX = MIN_MAX, SCORE_TOLERANCE = 1e-5) # decide on min/max problem and then run!
        model.set_params(**best_param) # update some parameters with the best so far
        print('Best Param = ',best_param)
        print(best_fval)    
        print(model)
    
    test_pred, valid_score = user_defined_eval_function(train, test, features, target, model, eval_metric, MIN_MAX, predict_test_output = True)
    toc = time.time()
    print('Time Elapsed = %s sec'%(toc - tic))
    print('Final Validatoin Score = ',valid_score)
    print(model) # final model confirmation
    
    return model, test_pred


# In[5]:


if __DEBUG_MODE__:
    import sys
    sys.path.append("../")
    import utils

    # test dataset
    if True:
        #--- dataset for classification ------------
        np.random.seed(seed = 123)

        NN = 1000 # the number of data points
        x1 = np.random.uniform(0, 1, NN)
        x2 = np.random.uniform(0, 1, NN)
        # y = 2*(x1 - 0.5) - (x2 - 0.5) > 0 # line
        # y = x1**2 - x2 > 0 # parabolic 1
        y = 2*x1*(1-x1) - x2 > 0 # parabolic 2
        # y = ((x1 - 0.5)**2 + (x2 - 0.5)**2 - 0.3**2) > 0 # circle

        # dataset
        df = pd.DataFrame({'x1':x1, 'x2':x2, 'y':y})
        for n in range(1,0): # add uncorrelated features for test
            df['r%s'%n] = np.random.uniform(0, 1, NN)

        # train, test
        train, test = train_test_split(df, test_size = 0.2)
        target = ['y']
        features = [f for f in df.columns if f not in target]
        print(len(features))
        features
    else:
        #--- dataset for regression ----------------
        np.random.seed(seed = 123)
        x_0 = np.linspace(0,  5, num=1000, endpoint=False); r_0 = np.random.normal(0, 0.3, len(x_0))
        x_1 = np.linspace(5, 10, num=1000, endpoint=False); r_1 = np.random.normal(0, 0.3, len(x_1))
        y0 = np.concatenate((np.sin(x_0)+5, np.sin(4*x_1)))
        # y0 = np.concatenate((np.sin(x_0), np.sin(1*x_1)))

        x = np.concatenate((x_0, x_1))
        y = y0 + np.concatenate((r_0, r_1))

        df = pd.DataFrame({'x':x, 'y0':y0, 'y':y})

        # train, test
        train, test = train_test_split(df, test_size = 0.2)
        features = ['x']
        target = ['y']

    #--- LigthGBM -----------------------------------------------------------------
    eval_metric = 'auc'; MIN_MAX = 'max'; lgbmodel = LGBMClassifier(learning_rate = 0.1, n_jobs = 4, random_state = 123) # classification 
    # eval_metric = 'rmse'; MIN_MAX = 'min'; lgbmodel = LGBMRegressor(learning_rate = 0.1, n_jobs = 4, random_state = 123) # regression

    params_schedule = [
        #ref) https://lightgbm.readthedocs.io/en/latest/Parameters-Tuning.html
        #ref) https://lightgbm.readthedocs.io/en/latest/Parameters.html
        #ref) https://www.reddit.com/r/MachineLearning/comments/aspx8x/d_methods_for_hyperparameter_tuning_with_lightgbm/
        {'learning_rate': ['log', [0.1]]},

        {'num_leaves': ['int', [8, 31, 128, 1000]],  # uni / log / int
    #      'min_child_samples': ['int', [0, 20, 200]]}, # min_data_in_leaf (uni / log / int)
#          'min_child_samples': ['int', [0, 10, 20, 50, 100, 200, 500, 1000]]}, # min_data_in_leaf (uni / log / int)
         'min_child_samples': ['int', [0, 10, 20, 50, 100, 200]]}, # min_data_in_leaf (uni / log / int)
    #          'min_child_weight': ['log', [0.001, 1000]]}, # min_sum_hessian_in_leaf (uni / log / int)
        {'min_split_gain': ['uni', [0, 1, 3, 10, 30, 100, 300, 1000]]}, #min_gain_to_split (uni / log / int)
        {'subsample': ['uni', [0.2,0.4,0.6,0.8,1.0]], # bagging_fraction
    #      'subsample_freq': ['int', [0,10,100]], # bagging_freq
         'colsample_bytree': ['uni', [0.2, 0.4, 0.6, 0.8, 1.0]]},  # feature_fraction (uni / log / int)
        {'reg_alpha': ['uni', [0, 1, 10, 100, 500]], # lambda_l1 
         'reg_lambda': ['uni', [0, 1, 10, 100, 500]]}, # lambda_l2 (uni / log / int)

        {'learning_rate': ['log', [0.005, 0.1, 0.2]]} # uni / log / int
    ]
    lgbmodel, test_pred_lgb = xgrid_search_boosting(train, test, features, target, params_schedule, lgbmodel, eval_metric, MIN_MAX,
                                CV_MODE = 'cross validation 1', N_FOLD = 5, N_BOOST_ROUND = 10000, EARLY_STOPPING = 50, RAND_SEED = 123)
    print('Train'); utils.eval_model_scores(lgbmodel, train, features, target)
    print('Test'); utils.eval_model_scores(lgbmodel, test, features, target)
    
    #--- XGBoost -----------------------------------------------------------------
    eval_metric = 'auc'; MIN_MAX = 'max'; xgbmodel = XGBClassifier(learning_rate = 0.1, n_jobs = 4, random_state = 123) # classification 
#     eval_metric = 'rmse'; MIN_MAX = 'min'; xgbmodel = XGBRegressor(learning_rate = 0.1, n_jobs = 4, random_state = 123) # regression

    params_schedule = [
#         #ref) https://www.analyticsvidhya.com/blog/2016/03/complete-guide-parameter-tuning-xgboost-with-codes-python/
#         {'learning_rate': ['log', [0.1]]},

#         {'max_depth': ['int', [3, 5, 7, 9, 11]], # uni / log / int
#          'min_child_weight': ['log', [1, 10, 100, 1000]]}, # uni / log / int
#         {'gamma': ['log', [0, 0.001, 1, 1000]]}, # uni / log / int
#         {'subsample': ['uni', [0.2, 0.4, 0.6, 0.8, 1.0]],  # uni / log / int
#          'colsample_bytree': ['uni', [0.2, 0.4, 0.6, 0.8, 1.0]]},  # uni / log / int
#         {'reg_alpha': ['log', [0, 0.001, 1, 1000]],
#          'reg_lambda': ['log', [1, 10, 100, 1000]]},  # uni / log / int

#         {'learning_rate': ['log', [0.005, 0.1, 0.2]]} # uni / log / int
    ]

    xgbmodel, test_pred_xgb = xgrid_search_boosting(train, test, features, target, params_schedule, xgbmodel, eval_metric, MIN_MAX,
                                CV_MODE = 'cross validation 1', N_FOLD = 5, N_BOOST_ROUND = 10000, EARLY_STOPPING = 50, RAND_SEED = 123)
    print('Train'); utils.eval_model_scores(xgbmodel, train, features, target)
    print('Test'); utils.eval_model_scores(xgbmodel, test, features, target)
    print(utils.feature_importance_xgboost(features, xgbmodel))


# In[ ]:





# In[ ]:







#!/usr/bin/env python
# coding: utf-8

# In[1]:


import numpy as np
import pandas as pd
import datetime

# import webdatareader_naver 
# import webdatareader_fred
from . import webdatareader_naver
from . import webdatareader_fred
__DEBUG_MODE__ = False


# In[2]:


def get_db(stocknames, varnames, SL_FILEPATH = './stock_list.csv', DB_DIRPATH = './project_stock_db_csv/'): # https://www.datacamp.com/community/tutorials/finance-python-trading
    sl = pd.read_csv(SL_FILEPATH, header=0) # stock list
    sl.index = sl.pop('stockname')
    #print(sl)
    def get_data(stockname):
        filepath = DB_DIRPATH + stockname + '.csv'
        #--- read data --------------------------------------
        df = pd.read_csv(filepath)
        #--- adjust types -----------------------------------
        df.columns = sl.loc[stockname]['colname'].split()
        df.index = pd.to_datetime(df.pop('Date'))
        for varname in varnames:
            if any(df.columns == varname):
                df[varname] = pd.to_numeric(df[varname]) # make sure variable is numeric
            else:
                df[varname] = np.nan
        return df[varnames]
    datas = map(get_data, stocknames)
    return (pd.concat(datas, keys=stocknames, names=['Ticker', 'Date']))
def update_db(stocknames, varnames, SL_FILEPATH = './stock_list.csv', DB_DIRPATH = './project_stock_db_csv/', print_url = False): # https://www.datacamp.com/community/tutorials/finance-python-trading
    sl = pd.read_csv(SL_FILEPATH, header=0) # stock list
    sl.index = sl.pop('stockname')
    #print(sl)
    def update_data(stockname):
        filepath = DB_DIRPATH + stockname + '.csv'
        #--- read data --------------------------------------
        if sl.loc[stockname]['source'] == 'naver': # naver website
            url_ = sl.loc[stockname]['link']
            df = webdatareader_naver.update_data_in_db(filepath, url_, print_url = print_url)
        elif sl.loc[stockname]['source'] == 'fred': # FRED website
            ticker = sl.loc[stockname]['ticker']
            df = webdatareader_fred.update_data_in_db(filepath, ticker)
        #--- adjust types -----------------------------------
        df.columns = sl.loc[stockname]['colname'].split()
        df.index = pd.to_datetime(df.pop('Date'))
        for varname in varnames:
            if any(df.columns == varname):
                df[varname] = pd.to_numeric(df[varname]) # make sure variable is numeric
            else:
                df[varname] = np.nan
        return df[varnames]
    datas = map(update_data, stocknames)
    return (pd.concat(datas, keys=stocknames, names=['Ticker', 'Date']))
def stock_split_weight(dates, stockname, SST_FILEPATH = './stock_split_table.csv'):
    df = pd.DataFrame(index = dates, data = {'Weight':1.0}) # initialize
    tables = pd.read_csv(SST_FILEPATH, header = 0) # load stock split table
    table = tables.loc[tables['stockname'] == stockname, ] # select data of a certail stockname
    table.index = pd.RangeIndex(len(table.index)) # reset the index: https://stackoverflow.com/questions/20490274/how-to-reset-index-in-a-pandas-data-frame
    print(table)
    for irow in range(table.shape[0]):
        Date = table.loc[irow, 'Date']
        Weight = table.loc[irow, 'Weight']
        df.loc[df.index <= Date, 'Weight'] *= Weight
    return df


# In[7]:


SL_FILEPATH = 'C:/Users/bomsoo/bomsoo1/project/projects/_stock_currency_prediction/stock_list.csv' # stock list file path
DB_DIRPATH = 'C:/Users/bomsoo/bomsoo1/project/projects/_stock_currency_prediction/project_stock_db_csv/' # database directory path
SST_FILEPATH = 'C:/Users/bomsoo/bomsoo1/project/projects/_stock_currency_prediction/stock_split_table.csv'

if __DEBUG_MODE__ and False:
#     df = update_db(
    df = get_db(
        ['Naver', 'SamsungElectronics', 'KRW', 'KOSPI', 'Gold',
         'unemployment_US', 'nonfarm_payroll_employment_US', 'Federal_Funds_Target_Rate', 'Foreign_Exchange_Rate_Korea_US', 
         'Foreign_Exchange_Rate_Canada_US', 'Consumer_Price_Index_for_All_Urban_Consumers_All_Items']
        , ['Price', 'Volume'], SL_FILEPATH, DB_DIRPATH)
    #----------------------------------------------------------------------
    price_df = df.reset_index().pivot('Date','Ticker','Price')
    price_df['unemployment_US'].dropna().plot(grid = True, figsize=(15,8))
    price_df['KRW'].dropna().plot(grid = True, secondary_y=True)

if __DEBUG_MODE__ and False:
    import matplotlib.pyplot as plt
    get_ipython().run_line_magic('matplotlib', 'inline')
    
#     df = update_db(['Naver', 'SamsungElectronics', 'KRW', 'KOSPI', 'Gold'], ['Price', 'Volume'], SL_FILEPATH, DB_DIRPATH)
    df = get_db(['Naver', 'SamsungElectronics', 'KRW', 'KOSPI', 'Gold'], ['Price', 'Volume'], SL_FILEPATH, DB_DIRPATH)
    #----------------------------------------------------------------------
    price_df = df.reset_index().pivot('Date','Ticker','Price')
    pct_change_price = price_df.pct_change()
    pct_change_price.hist(bins=400, sharex=True, figsize=(12,5))
    plt.show()
    print(pct_change_price.describe())
    #----------------------------------------------------------------------
    pd.scatter_matrix(pct_change_price, diagonal='kde', alpha=0.1,figsize=(12,12))
    plt.show()
    #--- moving average -------------------------------------------------------------------
    naver = df.loc[['Naver'],['Price']]
    naver['40'] = naver['Price'].rolling(window=40).mean()
    naver['252'] = naver['Price'].rolling(window=252).mean()
    print(naver.head())
    naver.plot()
    plt.show()
    #----------------------------------------------------------------------
    cum_daily_return = (1 + pct_change_price).cumprod()
    cum_daily_return.plot(figsize=(12,8))
    plt.show()
    cum_daily_return.tail()

    #mon_return = pct_change_price.resample('M').apply(lambda x: x[-1])
    mon_return = pct_change_price.resample('M').mean()
    mon_return.tail()
    
if __DEBUG_MODE__ and False:
    tickername = 'Naver'
    df = get_db([tickername], ['Price', 'Volume'], SL_FILEPATH, DB_DIRPATH).loc[tickername] # strip out the tickername in the row index
    df['Price_ori'] = df['Price'] # back-up
    df['Weight'] = stock_split_weight(df.index, tickername, SST_FILEPATH)
    df['Price'] = df['Price'] / df['Weight'] # correction
    df[['Price','Price_ori']].plot(figsize=(15,5), grid=True, color=['r','k'])


# In[ ]:








# coding: utf-8

# In[1]:


import numpy as np
import matplotlib.pyplot as plt
import time
import math
#-------------------------------
import keras
from keras.models import Sequential
from keras.layers import Dense, Activation, BatchNormalization
from keras.optimizers import SGD, Adam


# In[2]:


def train(x_train, y_train, x_test, y_test, hpar):
    input_shape = x_train.shape[1:]
    output_num = y_train.shape[1]
    if hpar['activation_last_layer'] == 'sigmoid': # loss function
        loss = 'binary_crossentropy'
    elif hpar['activation_last_layer'] == 'softmax':
        loss = 'categorical_crossentropy'
    if hpar['optimizer'] == 'gradient descent': # optimizer
        optimizer = SGD(lr=hpar['learning_rate'])
    elif hpar['optimizer'] == 'adam':
        optimizer = Adam(lr=hpar['learning_rate'])
    #-----------------------------------------------------
    model = Sequential()
    model.add(BatchNormalization(input_shape=input_shape))
    for i in range(len(hpar['hidden_layer_dims'])):
        model.add(Dense(hpar['hidden_layer_dims'][i]))
        model.add(BatchNormalization())
        model.add(Activation(hpar['activation_hidden_layers'])) # 'sigmoid', 'tanh', 'relu'
    model.add(Dense(output_num))
    model.add(BatchNormalization())
    model.add(Activation(hpar['activation_last_layer']))
    model.compile(optimizer = optimizer, loss = loss, metrics = ['accuracy'])
    #-----------------------------------------------------
    model.summary()
    loss = []
    accu = []
    for i in range(1, 1 + max(1, hpar['num_epochs'] // hpar['print_cost'])):
        tic = time.time()
        fout = model.fit(x_train, y_train, epochs=hpar['print_cost'], batch_size=hpar['minibatch_size'], class_weight=hpar['class_weight'], verbose=0)
        toc = time.time()
        loss = loss + fout.history['loss']
        accu = accu + fout.history['acc']
        score_train = model.evaluate(x_train, y_train, batch_size=hpar['minibatch_size'], verbose=0)
        score_test = model.evaluate(x_test, y_test, batch_size=hpar['minibatch_size'], verbose=0)
        print('>> epochs = ', (i*hpar['print_cost']),'; Train loss/acc = %f/%f'%(score_train[0],score_train[1]), '; Test loss/acc = %f/%f'%(score_test[0],score_test[1]), '; time = %f sec '%(toc-tic))
    mout = {'model': model,
            'loss': loss,
            'accu': accu}
    return mout


# In[3]:


def generate_2D_dataset(num_sample = 1000, seed = 1):
    np.random.seed(seed)
    X = np.random.rand(2, num_sample)
    #Y = np.int32((X[1,:] - X[0,:] + 0.3) > 0).reshape(1, num_sample)  # line
    #Y = np.int32((X[0,:]**2 + X[1,:]**2 - 0.7**2) > 0).reshape(1, num_sample) # semi-circle
    #Y = np.int32(((X[0,:]-0.5)**2 + (X[1,:]-0.5)**2 - 0.3**2) > 0).reshape(1, num_sample)  # circle
    #Y = np.int32((((X[0,:]-0.5)/0.4)**2 + ((X[1,:]-0.2)/0.05)**2 - 1) > 0).reshape(1, num_sample)  # ellipse
    Y = np.int32(((((X[0,:]-0.5)/0.4)**2 + ((X[1,:]-0.2)/0.05)**2 - 1) > 0) & (((X[0,:]-0.7)**2 + (X[1,:]-0.7)**2 - 0.1**2) > 0)).reshape(1, num_sample) # ellipse & circle
    #-----------------------------------------------------------------
    rn = np.random.permutation(num_sample)
    NofT = math.floor(num_sample * 0.7) # number of train samples
    X_train = X[:,rn[0:NofT]]
    Y_train = Y[:,rn[0:NofT]]
    X_test = X[:,rn[NofT:]]
    Y_test = Y[:,rn[NofT:]]
    return X_train, Y_train, X_test, Y_test
def convert_keras_input(X_train, Y_train, X_test, Y_test):
    x_train = X_train.T
    y_train = Y_train.T
    x_test = X_test.T
    y_test = Y_test.T
    y_train_one_hot = keras.utils.to_categorical(y_train, num_classes=2)
    y_test_one_hot = keras.utils.to_categorical(y_test, num_classes=2)
    print("x_train = ", x_train.shape)
    print("y_train = ", y_train.shape)
    print("x_test = ", x_test.shape)
    print("y_test = ", y_test.shape)
    print("y_train_one_hot = ", y_train_one_hot.shape)
    print("y_test_one_hot = ", y_test_one_hot.shape)
    return x_train, y_train, x_test, y_test, y_train_one_hot, y_test_one_hot
if False:
    X_train, Y_train, X_test, Y_test = generate_2D_dataset()
    x_train, y_train, x_test, y_test, y_train_one_hot, y_test_one_hot = convert_keras_input(X_train, Y_train, X_test, Y_test)
    #-------------------------------------------------------------
    hpar = {}
    hpar['hidden_layer_dims'] = [20,10,5]
    hpar['activation_hidden_layers'] = ['sigmoid', 'tanh', 'relu'][2]
    hpar['activation_last_layer'] = ['sigmoid', 'softmax'][1]
    hpar['learning_rate'] = 0.01
    hpar['num_epochs'] = 3000
    hpar['minibatch_size'] = 999999
    hpar['optimizer'] = ["gradient descent", "momentum", "adam"][2]
    hpar['class_weight'] = None
    hpar['print_cost'] = 1000 # print epoch step; no print when it is zero
    mout = train(x_train, y_train_one_hot, x_test, y_test_one_hot, hpar)
    #-------------------------------------------------------------
    plt.plot(mout['loss'])
    plt.show()
    plt.plot(mout['accu'])
    plt.show()
    #-------------------------------------------------------------
    yp_test_one_hot = mout['model'].predict(x_test)
    yp_test = np.argmax(yp_test_one_hot, axis = 1).reshape(y_test.shape)
    accuracy_test = np.mean(y_test == yp_test)
    print(accuracy_test)  


# In[ ]:







__version__ = '2019-08-08 12:00AM'

### data statistics ############################################
import numpy as np
import pandas as pd

def Data_Stat(df, n_row_view = 3):
    types = pd.Series({c:str(df[c].dtype) for c in df.columns}) # variable type
    total = df.isnull().sum() # the number of null values
    percent = 100 * total / df.isnull().count() # the percentage of null values
    n_unique = df.nunique() # the number of unique values
    
    tt = np.transpose(pd.concat([types, total, percent, n_unique], axis=1, keys=['Types', '# of nulls', '% of nulls', '# of uniques']))
    return pd.concat([df.head(n_row_view), tt], axis = 0)

### Evaludate Model Scores ###################################
import numpy as np
from sklearn.metrics import accuracy_score
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import f1_score
from sklearn.metrics import roc_auc_score
from sklearn.metrics import mean_squared_error

def Eval_Model_Scores(model, data, features, target, P_THRESHOLD = 0.5):
    num_cases = data.shape[0]; print('Total number of observations = %s'%num_cases)
    if ('Classifier' in str(model)) or ('LogisticRegression' in str(model)) or ('SVC' in str(model)):
        y_pred_prob = model.predict_proba(data[features])[:,1]
        y_pred = y_pred_prob >= P_THRESHOLD
        
        accuracy = accuracy_score(data[target], y_pred); print('Accuracy = %s'%accuracy)
        precision = precision_score(data[target], y_pred); print('Precision = %s'%precision)
        recall = recall_score(data[target], y_pred); print('Recall = %s'%recall)
        f1score = f1_score(data[target], y_pred); print('F1_Score = %s'%f1score)
        auc = roc_auc_score(data[target], y_pred_prob); print('AUC = %s'%auc)

        df_confusion = pd.crosstab(data[target].values.ravel(), y_pred, rownames=['Actual'], colnames=['Predicted']) # , margins=True)
        
        scores = (accuracy, precision, recall, f1score, auc, df_confusion, y_pred_prob)
    if 'Regressor' in str(model):
        y_pred = model.predict(data[features])

        mse = mean_squared_error(data[target], y_pred); print('MSE = %s'%mse)
        rmse = np.sqrt(mse); print('RMSE = %s'%rmse)

        scores = (mse, rmse, y_pred)
        
    return scores

### plot decision boundary for dataset with two feature variables #################
import matplotlib.pyplot as plt
import seaborn as sns

def Plot_2D_Decision_Bourndary(model, train, test, features, target):
    # update: 07/21/2019
    # ref) https://stackoverflow.com/questions/22294241/plotting-a-decision-boundary-separating-2-classes-using-matplotlibs-pyplot
    ft1 = features[0]; ft2 = features[1]

    # X, Y, Z for contour plot
    x0 = np.linspace(train[ft1].min(),  train[ft1].max(), num=1000)
    y0 = np.linspace(train[ft2].min(),  train[ft2].max(), num=1000)
    X, Y = np.meshgrid(x0, y0)
    zz = model.predict(pd.DataFrame({ft1:X.ravel(), ft2:Y.ravel()})) # Predict the transformed test documents
    Z = np.reshape(zz, X.shape)

    # plot
    fig, ax = plt.subplots(ncols=2, sharey=True, figsize=(12, 6))
    sns.scatterplot(x=ft1, y=ft2, data=train, hue=target[0], ax=ax[0]) # plot train dataset
    ax[0].contour(X, Y, Z, colors = 'black') # contour plot
    ax[0].legend(loc='upper right')
    sns.scatterplot(x=ft1, y=ft2, data=test, hue=target[0], ax=ax[1]) # plot test dataset
    ax[1].contour(X, Y, Z, colors = 'black') # contour plot
    ax[1].legend(loc='upper right')
    plt.show()
    
### feature importance for feature selection ##############################
import pandas as pd

def Feature_Importance_xgboost(features, xgmodel):
    # https://towardsdatascience.com/be-careful-when-interpreting-your-features-importance-in-xgboost-6e16132588e7
    xgmodel.get_booster().get_score(importance_type='gain') 
    # xgmodel.feature_importances_
    fi = pd.DataFrame(list(zip(features, xgmodel.feature_importances_)), columns = ['feature','importance']) # https://datascience.stackexchange.com/questions/48330/how-to-get-xgbregressor-feature-importance-by-column-name
    fi_sort = fi.sort_values(by='importance', ascending = False).reset_index(drop=True)
    fi_sort['cumsum'] = fi_sort['importance'].cumsum()
    return fi_sort

### Plot Hyperopt Resuls: (Score vs. Parameters), (Params vs. Time) ###########################
from hyperopt import Trials
import matplotlib.pyplot as plt

def Plot_Hyperopt_Score_vs_Params(param_names, trials): #ref) https://www.districtdatalabs.com/parameter-tuning-with-hyperopt
    f, axes = plt.subplots(nrows=1, ncols=len(param_names), figsize=(15,5))
    cmap = plt.cm.jet
    for i, val in enumerate(param_names):
        xs = [t['misc']['vals'][val] for t in trials.trials]
        ys = [-t['result']['loss'] for t in trials.trials]
    #     xs, ys = zip(sorted(zip(xs, ys)))
        ys = np.array(ys)
        if len(param_names) > 1:
            axes[i].scatter(xs, ys, s=20, linewidth=0.01, alpha=0.75, c=cmap(float(i)/len(param_names)))
            axes[i].set_xlabel(val, fontsize=12)
            axes[i].set_ylabel('cross validation accuracy', fontsize=12)
        else:
            axes.scatter(xs, ys, s=20, linewidth=0.01, alpha=0.75, c=cmap(float(i)/len(param_names)))
            axes.set_title(val)
            axes.set_xlabel(val, fontsize=12)
            axes.set_ylabel('cross validation accuracy', fontsize=12)

def Plot_Hyperopt_Params_vs_Time(param_names, trials):
    f, axes = plt.subplots(nrows=1, ncols=len(param_names), figsize=(15,5))
    cmap = plt.cm.jet
    for i, val in enumerate(param_names):
        xs = [t['tid'] for t in trials.trials]
        ys = [t['misc']['vals'][val] for t in trials.trials]
        if len(param_names) > 1:
            axes[i].set_xlim(xs[0]-10, xs[-1]+10)
            axes[i].scatter(xs, ys, s=20, linewidth=0.01, alpha=0.75, c=cmap(float(i)/len(param_names)))
            axes[i].set_xlabel('$t$', fontsize=16)
            axes[i].set_ylabel(val, fontsize=16)  
        else:
            axes.set_xlim(xs[0]-10, xs[-1]+10)
            axes.scatter(xs, ys, s=20, linewidth=0.01, alpha=0.75, c=cmap(float(i)/len(param_names)))
            axes.set_xlabel('$t$', fontsize=16)
            axes.set_ylabel(val, fontsize=16)  

### Find the Pattern of Characters #################################################            
# update 2019-08-07
def Find_Characters_Pattern(string, DICT_CH_TO_PATTERN = {
    'a':'A', 'b':'A', 'c':'A', 'd':'A', 'e':'A', 'f':'A', 'g':'A', 'h':'A', 'i':'A', 'j':'A', 'k':'A', 'l':'A', 'm':'A', 
    'n':'A', 'o':'A', 'p':'A', 'q':'A', 'r':'A', 's':'A', 't':'A', 'u':'A', 'v':'A', 'w':'A', 'x':'A', 'y':'A', 'z':'A',
    '0':'D', '1':'D', '2':'D', '3':'D', '4':'D', '5':'D', '6':'D', '7':'D', '8':'D', '9':'D',
}):
    dkeys = DICT_CH_TO_PATTERN.keys()
    pt = ''.join([DICT_CH_TO_PATTERN[c] if c in dkeys else c for c in string.lower()])
    return ''.join([c for i, c in enumerate(pt) if (i==0) | ((i != 0) & (pt[i-1] != pt[i]))])
# Find_Characters_Pattern('234 AB - 4 ((77B))??//') # 'D A - D (DA)?/'            


#!/usr/bin/env python
# coding: utf-8

# In[1]:


import pandas as pd


# In[2]:


#ref) https://medium.com/@ageitgey/quick-tip-the-easiest-way-to-grab-data-out-of-a-web-page-in-python-7153cecfca58
def download_table(url): 
    dfs = pd.read_html(url, header=0) # read tables and save them into a list of dataframes
    df = dfs[0]  # select the first element of the list
    df.dropna(axis=0, inplace=True) # axis=0: remove rows that contain NaN
    ii = df.iloc[:,0].str.contains('\d').values # excluding non-digit rows by checking the first column
    return df.loc[ii,]
def download_tables(url_, start=1, end=5, print_url = False): # read table data from the start page to end page
    dfs = list()
    for i in range(start, end+1):
        url = url_ + str(i) # create a complete url for naver web page
        if (print_url == True):
            print(url)
        df0 = download_table(url) # download a table from a naver web page
        if (i != start) and (pd.DataFrame.equals(dfs[-1], df0) or len(df0.index)==0): # stop conditions
            break
        else:
            dfs.append(df0)
    df = pd.concat(dfs)
    return df


# In[3]:


def update_data_in_db(filepath, url_, COL_DATE = '날짜', MAX_PAGE = 99999, print_url = False):
    try:
        df = pd.read_csv(filepath) # read the already sorted data
        latestDate = df.iloc[-1][COL_DATE] # temporarily save the latest date
        for i in range(1, MAX_PAGE):
            url = url_ + str(i) # create a complete url for naver web page
            df0 = download_table(url)
            if any(df0[COL_DATE] == latestDate):
                df.drop(df.index[df[COL_DATE] == latestDate], inplace=True) # drop the last row 
                df = pd.concat([df, df0[df0[COL_DATE] >= latestDate]])
                break
            else:
                df = pd.concat([df, df0])
    except:
        print("File not found in the database. Now, searching the internet to download information...")
        df = download_tables(url_, start = 1, end = MAX_PAGE, print_url = print_url) # download the data from the website
        print("Downloading completed.")
    df.sort_values(COL_DATE, inplace = True) # sort in order of date & time
    df.to_csv(filepath, index=False) # save the data to db
    return df


# In[6]:


if False:
#     df = download_table('http://finance.naver.com/item/sise_day.nhn?code=035420&page=1')
#     df = download_table('http://info.finance.naver.com/marketindex/exchangeDailyQuote.nhn?marketindexCd=FX_USDKRW&page=1')
    df = download_table('https://finance.naver.com/marketindex/exchangeDailyQuote.nhn?marketindexCd=FX_CADKRW&page=403')
    print(df)
if False:
#     df = download_tables('http://finance.naver.com/item/sise_day.nhn?code=035420&page=', 1, 15)
#     df = download_tables('https://finance.naver.com/item/sise_day.nhn?code=000660&page=', 542, 547, print_url = True)
    df = download_tables('https://finance.naver.com/marketindex/exchangeDailyQuote.nhn?marketindexCd=FX_CADKRW&page=', 376, 380, print_url = True)
    print(df)
if False:
#    df = update_data_in_db('c:/Users/bomso/bomsoo1/project/python/naver.csv', 'http://finance.naver.com/item/sise_day.nhn?code=035420&page=')
#    df = update_data_in_db('c:/Users/bomso/bomsoo1/project/python/samsung.csv', 'http://info.finance.naver.com/marketindex/exchangeDailyQuote.nhn?marketindexCd=FX_USDKRW&page=')
#    df = update_data_in_db('c:/Users/bomso/bomsoo1/project/python/kospi.csv', 'http://finance.naver.com/sise/sise_index_day.nhn?code=KOSPI&page=')
#     df = update_data_in_db('c:/Users/bomso/bomsoo1/project/python/gold.csv', 'http://info.finance.naver.com/marketindex/goldDailyQuote.nhn?&page=')
    df = update_data_in_db('C:/Users/bomsoo/bomsoo1/project/python/project_stock_db_csv/CADKRW.csv', 'https://finance.naver.com/marketindex/exchangeDailyQuote.nhn?marketindexCd=FX_CADKRW&page=')
    
    #df['종가'].plot(xlim=(pd.Timestamp('2016-01-01'), pd.Timestamp.today()), ylim=(400000,1000000))

    print(df.iloc[-35:])


# In[ ]:





# In[ ]:








# coding: utf-8

# In[1]:


import numpy as np
import time
import pandas as pd
np.random.seed(0)
from keras.models import Model
from keras.layers import Dense, Input, Dropout, LSTM, Activation
from keras.layers.embeddings import Embedding
from keras.preprocessing import sequence
from keras.initializers import glorot_uniform
np.random.seed(1)


# In[2]:


def pretrained_embedding_layer(word_to_vec_map, word_to_index):
    """
    Creates a Keras Embedding() layer and loads in pre-trained GloVe 50-dimensional vectors.
    
    Arguments:
    word_to_vec_map -- dictionary mapping words to their GloVe vector representation.
    word_to_index -- dictionary mapping from words to their indices in the vocabulary (400,001 words)

    Returns:
    embedding_layer -- pretrained layer Keras instance
    """
    vocab_len = len(word_to_index) + 1                  # adding 1 to fit Keras embedding (requirement)
    emb_dim = word_to_vec_map["cucumber"].shape[0]      # define dimensionality of your GloVe word vectors (= 50)
    emb_matrix = np.zeros((vocab_len, emb_dim)) # Initialize the embedding matrix as a numpy array of zeros of shape (vocab_len, dimensions of word vectors = emb_dim)
    for word, index in word_to_index.items(): # Set each row "index" of the embedding matrix to be the word vector representation of the "index"th word of the vocabulary
        emb_matrix[index, :] = word_to_vec_map[word]
    embedding_layer = Embedding(vocab_len, emb_dim, trainable = False) # Define Keras embedding layer with the correct output/input sizes, make it trainable. Use Embedding(...). Make sure to set trainable=False. 
    embedding_layer.build((None,)) # Build the embedding layer, it is required before setting the weights of the embedding layer. Do not modify the "None".
    embedding_layer.set_weights([emb_matrix]) # Set the weights of the embedding layer to the embedding matrix. Your layer is now pretrained.
    return embedding_layer


# In[3]:


def sentences_to_indices(X, word_to_index, max_len):
    """
    Converts an array of sentences (strings) into an array of indices corresponding to words in the sentences.
    The output shape should be such that it can be given to `Embedding()` (described in Figure 4). 
    
    Arguments:
    X -- array of sentences (strings), of shape (m, 1)
    word_to_index -- a dictionary containing the each word mapped to its index
    max_len -- maximum number of words in a sentence. You can assume every sentence in X is no longer than this. 
    
    Returns:
    X_indices -- array of indices corresponding to words in the sentences from X, of shape (m, max_len)
    """
    m = X.shape[0]                                   # number of training examples
    X_indices = np.zeros((m, max_len)) # Initialize X_indices as a numpy matrix of zeros and the correct shape (≈ 1 line)
    for i in range(m):                               # loop over training examples
        sentence_words = X[i].lower().split() # Convert the ith training sentence in lower case and split is into words. You should get a list of words.
        j = 0 # Initialize j to 0
        for w in sentence_words: # Loop over the words of sentence_words
            X_indices[i, j] = word_to_index[w] # Set the (i,j)th entry of X_indices to the index of the correct word.
            j = j + 1 # Increment j to j + 1
    return X_indices
def train(word_to_vec_map, word_to_index, X_train_indices, Y_train_oh, X_test_indices, Y_test_oh, hpar):
    """
    Arguments:
    input_shape -- shape of the input, usually (max_len,)
    word_to_vec_map -- dictionary mapping every word in a vocabulary into its 50-dimensional vector representation
    word_to_index -- dictionary mapping from words to their indices in the vocabulary (400,001 words)

    Returns:
    model -- a model instance in Keras
    """
    input_shape = X_train_indices.shape[1:] # exclude the nubmer of samples
    output_num = Y_train_oh.shape[-1]
    embedding_layer = pretrained_embedding_layer(word_to_vec_map, word_to_index) # Create the embedding layer pretrained with GloVe Vectors (≈1 line)
    #-----------------------------------------
    sentence_indices = Input(shape=input_shape, dtype='int32') # Define sentence_indices as the input of the graph, it should be of shape input_shape and dtype 'int32' (as it contains indices).
    X = embedding_layer(sentence_indices) # Propagate sentence_indices through your embedding layer, you get back the embeddings
    for i in range(0, len(hpar['layer_lstm_hidden_state_dim_at_middle'])):
        X = LSTM(hpar['layer_lstm_hidden_state_dim_at_middle'][i], return_sequences=True)(X) # Propagate the embeddings through an LSTM layer with 128-dimensional hidden state # Be careful, the returned output should be a batch of sequences.
        X = Dropout(hpar['layer_dropout_rate_at_middle'][i])(X) # Add dropout with a probability of 0.5
    X = LSTM(hpar['layer_lstm_hidden_state_dim_at_output'])(X)# Propagate X trough another LSTM layer with 128-dimensional hidden state # Be careful, the returned output should be a single hidden state, not a batch of sequences.
    X = Dropout(hpar['layer_dropout_rate_at_output'])(X) # Add dropout with a probability of 0.5
#    X = Dense(5, activation='softmax')(X) # Propagate X through a Dense layer with softmax activation to get back a batch of 5-dimensional vectors.
    X = Dense(output_num)(X) # Propagate X through a Dense layer with softmax activation to get back a batch of 5-dimensional vectors.
    X = Activation('softmax')(X) # Add a softmax activation
    model = Model(inputs = sentence_indices, outputs = X) # Create Model instance which converts sentence_indices into X.
    #-----------------------------------------
    model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
    model.summary()
    log_df = pd.DataFrame(np.nan, index=range(hpar['num_epochs']), columns = ['loss', 'acc', 'loss_train', 'acc_train', 'loss_test', 'acc_test'])
    for i in range(1, 1 + max(1, hpar['num_epochs'] // hpar['print_cost'])):
        tic = time.time()
        fout = model.fit(X_train_indices, Y_train_oh, epochs=hpar['print_cost'], batch_size=hpar['minibatch_size'], shuffle=True, verbose=0)
        toc = time.time()
        log_df['loss'].loc[range((i-1)*hpar['print_cost'],i*hpar['print_cost'])] = fout.history['loss']
        log_df['acc'].loc[range((i-1)*hpar['print_cost'],i*hpar['print_cost'])] = fout.history['acc']
        loss_train, acc_train = model.evaluate(X_train_indices, Y_train_oh, verbose=0) # score = [loss, acc]
        loss_test, acc_test = model.evaluate(X_test_indices, Y_test_oh, verbose=0) # score = [loss, acc]
        log_df['loss_train'].loc[i*hpar['print_cost'] - 1] = loss_train
        log_df['acc_train'].loc[i*hpar['print_cost'] - 1] = acc_train
        log_df['loss_test'].loc[i*hpar['print_cost'] - 1] = loss_test
        log_df['acc_test'].loc[i*hpar['print_cost'] - 1] = acc_test
        print('>> epochs = ', (i*hpar['print_cost']),'; Train loss/acc = %f/%f'%(loss_train, acc_train), '; Test loss/acc = %f/%f'%(loss_test, acc_test), '; time = %f sec '%(toc-tic))
    mout = {'model': model,
            'log': log_df}
    return mout


# In[4]:


if False:
    import csv
    import emoji
    import matplotlib.pyplot as plt

    def read_glove_vecs(glove_file):
    #    with open(glove_file, 'r') as f:
        with open(glove_file, 'r', encoding="utf8") as f: # Bomsoo modified : https://stackoverflow.com/questions/9233027/unicodedecodeerror-charmap-codec-cant-decode-byte-x-in-position-y-character
            words = set()
            word_to_vec_map = {}
            for line in f:
                line = line.strip().split()
                curr_word = line[0]
                words.add(curr_word)
                word_to_vec_map[curr_word] = np.array(line[1:], dtype=np.float64)
        
            i = 1
            words_to_index = {}
            index_to_words = {}
            for w in sorted(words):
                words_to_index[w] = i
                index_to_words[i] = w
                i = i + 1
        return words_to_index, index_to_words, word_to_vec_map
    def read_csv(filename = 'data/emojify_data.csv'):
        phrase = []
        emoji = []
        with open (filename) as csvDataFile:
            csvReader = csv.reader(csvDataFile)

            for row in csvReader:
                phrase.append(row[0])
                emoji.append(row[1])
        X = np.asarray(phrase)
        Y = np.asarray(emoji, dtype=int)

        return X, Y
    emoji_dictionary = {"0": "\u2764\uFE0F",    # :heart: prints a black instead of red heart depending on the font
                        "1": ":baseball:",
                        "2": ":smile:",
                        "3": ":disappointed:",
                        "4": ":fork_and_knife:"}
    def label_to_emoji(label):
        """
        Converts a label (int or string) into the corresponding emoji code (string) ready to be printed
        """
        return emoji.emojize(emoji_dictionary[str(label)], use_aliases=True)

    # loading datasets
    word_to_index, index_to_word, word_to_vec_map = read_glove_vecs('../datasets/coursera_dnn5_week2_Emojify/data/glove.6B.50d.txt')
    X_train, Y_train = read_csv('../datasets/coursera_dnn5_week2_Emojify/data/train_emoji.csv')
    X_test, Y_test = read_csv('../datasets/coursera_dnn5_week2_Emojify/data/tesss.csv')

    # converting datasets
    maxLen = len(max(X_train, key=len).split())
    X_train_indices = sentences_to_indices(X_train, word_to_index, maxLen)
    Y_train_oh = np.eye(5)[Y_train] # convert_to_one_hot
    X_test_indices = sentences_to_indices(X_test, word_to_index, max_len = maxLen)
    Y_test_oh = np.eye(5)[Y_test] # convert_to_one_hot 

    # training
    hpar = {}
    hpar['layer_lstm_hidden_state_dim_at_middle'] = [128, 128, 128] # list
    hpar['layer_dropout_rate_at_middle'] = [0.5, 0.5, 0.5] # list
    hpar['layer_lstm_hidden_state_dim_at_output'] = 128
    hpar['layer_dropout_rate_at_output'] = 0.5
    hpar['num_epochs'] = 100
    hpar['print_cost'] = 10
    hpar['minibatch_size'] = 32
    mout = train(word_to_vec_map, word_to_index, X_train_indices, Y_train_oh, X_test_indices, Y_test_oh, hpar)
    mout['log'].plot().legend(bbox_to_anchor=(1, 1)) # plot(marker='o')
    plt.show()
    
    # This code allows you to see the mislabelled examples
    pred = mout['model'].predict(X_test_indices)
    for i in range(len(X_test)):
        x = X_test_indices
        num = np.argmax(pred[i])
        if(num != Y_test[i]):
            print('Expected emoji:'+ label_to_emoji(Y_test[i]) + ' prediction: '+ X_test[i] + label_to_emoji(num).strip())    
            
    # Change the sentence below to see your prediction. Make sure all the words are in the Glove embeddings.  
    x_test = np.array(['not feeling happy'])
    x_test_indices = sentences_to_indices(x_test, word_to_index, maxLen)
    print(x_test[0] +' '+  label_to_emoji(np.argmax(mout['model'].predict(x_test_indices))))    




#!/usr/bin/env python
# coding: utf-8

# In[1]:


import numpy as np
__DEBUG_MODE__ = False


# In[2]:


nLIGHT = 255
nDARK = 125
nOBJECT = 0


# In[3]:


def light_side_of(M0, direction, nLIGHT, nOBJECT):
    M = np.copy(M0) # hard copy the input 2-D matrix, M0
    if direction == "left":
        lsource = np.full(M.shape[0], True) # set light source
        for n in range(M.shape[1]): # sweep light source from left to right
            ii = M[:,n] == nOBJECT
            lsource[ii] = False   # when meet objects, turn off light source
            M[lsource,n] = nLIGHT
    elif direction == "right":
        lsource = np.full(M.shape[0], True) # set light source
        for n in reversed(range(M.shape[1])): # sweep light source from right to left
            ii = M[:,n] == nOBJECT
            lsource[ii] = False  # when meet objects, turn off light source
            M[lsource,n] = nLIGHT
    elif direction == "bottom":
        lsource = np.full(M.shape[1], True) # set light source
        for n in range(M.shape[0]): # sweep light source from bottom to top
            ii = M[n,:] == nOBJECT
            lsource[ii] = False  # when meet objects, turn off light source
            M[n,lsource] = nLIGHT
    elif direction == "top":
        lsource = np.full(M.shape[1], True) # set light source
        for n in reversed(range(M.shape[0])): # sweep light source from top to bottom
            ii = M[n,:] == nOBJECT
            lsource[ii] = False  # when meet objects, turn off light source
            M[n,lsource] = nLIGHT
    return M
def data2matrix(data, data_lower_bound, data_upper_bound, data_resolution = 100):
    assert(len(data.shape) == 1) # assert that data be rank 0
    Nx = data.shape[0] # the number of data points
    Ny = data_resolution # the number of y-axis resolution
    nx = np.arange(0, Nx) # column indices for matrix M
    ny = ((data - data_lower_bound) / (data_upper_bound - data_lower_bound) * (Ny-1)).astype(int) # row indices for matrix M
    #---------------------------------------------------------------
    M = np.full((Ny, Nx), nDARK, dtype=np.uint8)
    M[ny,nx] = nOBJECT  # draw data points on the matrix
    #---------------------------------------------------------------
    MM = np.full((Ny, 2*Nx-1), nDARK, dtype=np.uint8) # expanded version of matrix
    MM[ny, np.arange(0,2*Nx-1,2)] = M[ny,nx]
    MM[ny[0:-1], np.arange(1,2*Nx-1,2)] = M[ny[0:-1],nx[0:-1]]
    MM[ny[1:], np.arange(1,2*Nx-1,2)] = M[ny[1:],nx[1:]]
    #---------------------------------------------------------------
    MM = light_side_of(MM, "top", nLIGHT, nOBJECT)
    MM = light_side_of(MM, "bottom", nLIGHT, nOBJECT)
    #---------------------------------------------------------------
    M[M == nDARK] = nLIGHT
    MM[MM == nDARK] = nOBJECT
    #---------------------------------------------------------------
    ddata = np.full(2*Nx-1, np.nan)
    ddata[np.arange(0,2*Nx-1,2)] = data[:]
    ddata[np.arange(1,2*Nx-1,2)] = (data[0:-1] + data[1:]) / 2.0
    #---------------------------------------------------------------
    data_scale = np.linspace(data_lower_bound, data_upper_bound, Ny, endpoint=True)
    return M, MM, ddata, data_scale
def find_valley_in_matrix(MM, ddata, data_scale, threshold, pourWaterAt = "top"):
    MM_lake = light_side_of(MM, pourWaterAt, nDARK, nOBJECT)  # fill with water
    MM_lake = light_side_of(MM_lake, "left", nLIGHT, nOBJECT) # spill over-flow to the left
    MM_lake = light_side_of(MM_lake, "right", nLIGHT, nOBJECT) # spill over-flow to the right
    #---------------------------------------------------------------
    TT = np.zeros(MM_lake.shape)
    TT[MM_lake == nDARK] = 1
    tt = np.sum(TT, axis = 0)
    #---------------------------------------------------------------
    d_scale = data_scale[1] - data_scale[0]
    percent_change = 100 * np.abs(d_scale * tt / ddata)
    #---------------------------------------------------------------
    ivalley = (percent_change == np.amax(percent_change)) & (percent_change > threshold)
    return ivalley, MM_lake
def find_valley_in_data(data, data_lower_bound, data_upper_bound, threshold, pourWaterAt = "top", data_resolution = 100):
    M, MM, ddata, data_scale = data2matrix(data, data_lower_bound, data_upper_bound, data_resolution)
    ivalley, MM_lake = find_valley_in_matrix(MM, ddata, data_scale, threshold, pourWaterAt)
    ilow = ivalley[np.arange(0,MM_lake.shape[1],2)]
    return ilow, M, MM, MM_lake
def find_valleys_in_data(data, data_lower_bound, data_upper_bound, threshold, pourWaterAt = "top", data_resolution = 100):
    N = data.shape[0]
    ilows = np.full(N, False)
    ilows[0] = True
    ilows[-1] = True
    nline = np.arange(0,N)
    icheck = np.full(N, True) # need-to-check area
    ifocus = np.full(N, True) # focus-to-check area
    for i in range(N):
        #print(i)
        ilow, M, MM, MM_lake = find_valley_in_data(data[ifocus], data_lower_bound, data_upper_bound, threshold, pourWaterAt, data_resolution)
        if any(ilow):
            ilows[ifocus] = ilows[ifocus] | ilow  # update lows
        else:
            icheck[nline[ifocus][0:-1]] = False
            #print(">> {} % completed...".format(100*(np.sum(~icheck)/len(icheck))))
        nums = nline[icheck & ilows]
        if len(nums) < 2:    # if there is no two points
            break
        ifocus = np.full(data.shape[0], False)
        ifocus[nums[0]:(nums[1]+1)] = True
    ilows[0] = False
    ilows[-1] = False
    return ilows, M, MM, MM_lake
def matrix2image(M):
    img = np.zeros((M.shape[0], M.shape[1], 3), dtype=np.uint8)
    img[:] = M.reshape(M.shape[0], M.shape[1], 1)
    return img


# In[4]:


def model(data, percent_threshold, data_resolution = 100):
    print("> Lows/Highs Detecter Algorithm started...")
    print(">> version = 2018-01-03")
    data_lower_bound = np.amin(data)
    data_upper_bound = np.amax(data)
    ilows, M1, MM1, MM_lake1 = find_valleys_in_data(data, data_lower_bound, data_upper_bound, percent_threshold, "top", data_resolution)
    percent_threshold2 = 100.0 * percent_threshold / (100.0 + percent_threshold) 
    ihighs, M2, MM2, MM_lake2 = find_valleys_in_data(data, data_lower_bound, data_upper_bound, percent_threshold2, "bottom", data_resolution)
    #plt.imshow(matrix2image(MM1))
    #plt.show()
    return ilows, ihighs
def find_nears(data, ilows, ihighs, percent_data_error = 0.5):
    assert(len(data.shape) == 1)
    iboth = ilows | ihighs
    nnn = np.arange(1, len(iboth)+1)
    nnl = np.insert(nnn[iboth], 0, 0) # insert zero to the beginning of the array
    nnr = np.append(nnn[iboth], len(iboth)+1) # insert zero to the end of the array
    rgl = np.diff(nnl) - 1 # number of left elements 
    rgr = np.diff(nnr) - 1 # number of right elements 
    #------------------------------------------------
    nbothl = np.zeros(iboth.shape, dtype=np.int64) # number of left elements
    nbothl[iboth] = -rgl
    nbothr = np.zeros(iboth.shape, dtype=np.int64) # number of right elements 
    nbothr[iboth] = +rgr
    #------------------------------------------------
    inearlows = np.copy(ilows) # near-lowest points
    inearhighs = np.copy(ihighs) # near-highest points
    for i in range(np.amin(nbothl), np.amax(nbothr)+1):
        if i > 0: # sweep for right elements
            n0 = np.arange(0, len(data)-i)
            n1 = np.arange(i, len(data))
        elif i < 0: # sweep for left elements
            n0 = np.arange(-i, len(data))
            n1 = np.arange(0, len(data)+i)
        if i != 0:
            ill = ilows[n0] & (nbothl[n0] <= i) & (i <= nbothr[n0]) & (abs((data[n1] - data[n0]) / data[n0]) <= (percent_data_error / 100.0))
            inearlows[n1] = inearlows[n1] | ill # update
            ihh = ihighs[n0] & (nbothl[n0] <= i) & (i <= nbothr[n0]) & (abs((data[n0] - data[n1]) / data[n1]) <= (percent_data_error / 100.0))
            inearhighs[n1] = inearhighs[n1] | ihh # update
    return inearlows, inearhighs


# In[5]:


if __DEBUG_MODE__:
    import matplotlib.pyplot as plt
    import time
    get_ipython().run_line_magic('matplotlib', 'inline')

    #np.random.seed(1)
    #data = np.random.rand(500) + 3 # example: random noise
    data = 1.0 + (5.1/100/2)*(1 + np.sin(np.linspace(0,42,1000))) # example: sine wave
    #---------------------------------------------------------------
    tic = time.time()
    ilows, ihighs = model(data, percent_threshold = 5, data_resolution = 100)   # threshold [%]
    inearlows, inearhighs = find_nears(data, ilows, ihighs, percent_data_error = 2) # error [%]
    toc = time.time()
    print(str(toc-tic) + " sec")
    #---------------------------------------------------------------
    plt.figure(figsize=(16,5))
    plt.plot(data,'k.-')
    plt.plot(np.arange(0,data.shape[0])[ilows], data[ilows],'ro', markersize=4)
    plt.plot(np.arange(0,data.shape[0])[ihighs], data[ihighs],'bo', markersize=4)
    plt.plot(np.arange(0,data.shape[0])[inearlows], data[inearlows],'ro', markersize=8, markerfacecolor='none')
    plt.plot(np.arange(0,data.shape[0])[inearhighs], data[inearhighs],'bo', markersize=8, markerfacecolor='none')
    #plt.xlim(800, 850)
    plt.show()


# In[ ]:








# coding: utf-8

# In[1]:


import tensorflow as tf
from tensorflow.python.framework import ops
import pickle
import numpy as np
import matplotlib.pyplot as plt
import time
get_ipython().magic('matplotlib inline')


# In[2]:


def random_mini_batches(X, Y, mini_batch_size = 64, seed = 0):
    np.random.seed(seed)            # To make your "random" minibatches the same as ours
    m = X.shape[1]                  # number of training examples
    mini_batches = []
    # Step 1: Shuffle (X, Y)
    permutation = np.random.permutation(m)
    shuffled_X = X[:, permutation]
    shuffled_Y = Y[:, permutation]
    # Step 2: Partition (shuffled_X, shuffled_Y). Minus the end case.
    num_complete_minibatches = max(1, m // mini_batch_size) # number of mini batches of size mini_batch_size in your partitionning
    for k in range(0, num_complete_minibatches):
        mini_batch_X = shuffled_X[:, (k*mini_batch_size):((k+1)*mini_batch_size)]
        mini_batch_Y = shuffled_Y[:, (k*mini_batch_size):((k+1)*mini_batch_size)]
        mini_batch = (mini_batch_X, mini_batch_Y)
        mini_batches.append(mini_batch)
    if (m >= mini_batch_size) and (m % mini_batch_size != 0): # Handling the end case (last mini-batch < mini_batch_size)
        mini_batch_X = shuffled_X[:, (num_complete_minibatches*mini_batch_size):]
        mini_batch_Y = shuffled_Y[:, (num_complete_minibatches*mini_batch_size):]
        mini_batch = (mini_batch_X, mini_batch_Y)
        mini_batches.append(mini_batch)
    return mini_batches
def create_placeholders(n_x, n_y):
    X = tf.placeholder(tf.float32, shape=(n_x,None), name='X')
    Y = tf.placeholder(tf.float32, shape=(n_y,None), name='Y')
    return X, Y
def initialize_parameters(layerdims, seed = 1):
    tf.set_random_seed(seed)                   # so that your "random" numbers match ours
    parameters = {}
    for i in range(1, len(layerdims)):
        parameters["W%d"%i] = tf.get_variable("W%d"%i, [layerdims[i],layerdims[i-1]], initializer = tf.contrib.layers.xavier_initializer(seed = seed))
        parameters["b%d"%i] = tf.get_variable("b%d"%i, [layerdims[i],1], initializer = tf.zeros_initializer())
    return parameters
def forward_propagation(X, parameters, hpar):
    L = len(parameters) // 2 # 2 -> there are two groups: W's and b's
    units = {}
    units['Z1'] = tf.add(tf.matmul(parameters['W1'],X),parameters['b1'])
    for i in range(2, L+1):
        if hpar['activation_hidden_layers'] == 'sigmoid':
            units['A'+str(i-1)] = tf.nn.sigmoid(units['Z'+str(i-1)]) # sigmoidfunction
        elif hpar['activation_hidden_layers'] == 'tanh':
            units['A'+str(i-1)] = tf.nn.tanh(units['Z'+str(i-1)]) # tanh function
        elif hpar['activation_hidden_layers'] == 'relu':
            units['A'+str(i-1)] = tf.nn.relu(units['Z'+str(i-1)]) # ReLU function
        units['Z'+str(i)] = tf.add(tf.matmul(parameters['W'+str(i)], units['A'+str(i-1)]), parameters['b'+str(i)])
    ZL = units['Z'+str(L)]
    return ZL # output of the last LINEAR unit
def compute_cost(ZL, Y, hpar):
    logits = tf.transpose(ZL) # to fit the tensorflow requirement for tf.nn.softmax_cross_entropy_with_logits(...,...)
    labels = tf.transpose(Y) # to fit the tensorflow requirement for tf.nn.softmax_cross_entropy_with_logits(...,...)
    if hpar['activation_last_layer'] == 'sigmoid':
        cost = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(logits = logits, labels = labels))
    elif hpar['activation_last_layer'] == 'softmax':
        cost = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(logits = logits, labels = labels))
    return cost
def normalize_X_Y(X_train0, Y_train0, X_test0, Y_test0, hpar):
    if hpar['input_normalization'] == True:
        Xu = np.mean(X_train0, axis = 1, keepdims = True)
        Xs = np.std(X_train0, axis = 1, keepdims = True)
    else:
        Xu = np.zeros((X_train0.shape[0],1))
        Xs = np.full((X_train0.shape[0],1), 1.0)
    X_train = (X_train0 - Xu) / Xs
    Y_train = np.int64(Y_train0 > 0.5)
    X_test = (X_test0 - Xu) / Xs
    if len(Y_test0) != 0:
        Y_test = np.int64(Y_test0 > 0.5)
    else:
        Y_test = []
    return X_train, Xu, Xs, Y_train, X_test, Y_test
def denormalize_W1_b1(WW1, bb1, Xu, Xs):
    W1 = WW1 / Xs.T
    b1 = bb1 - np.dot(WW1, Xu / Xs)
    return W1, b1


# In[3]:


def convert_to_one_hot(labels, C):
    C = tf.constant(C, name='C') # Create a tf.constant equal to C (depth), name it 'C'. (approx. 1 line)
    one_hot_matrix = tf.one_hot(labels.squeeze(), C, axis=0)
    sess = tf.Session()
    one_hot = sess.run(one_hot_matrix)
    sess.close()
    return one_hot
def train(X_train0, Y_train0, X_test0, Y_test0, hpar):
    print('> Deep Neural Network (tensorflow) started ...')
    X_train, Xu, Xs, Y_train, X_test, Y_test = normalize_X_Y(X_train0, Y_train0, X_test0, Y_test0, hpar)
    print('>> X_train shape = '+str(X_train.shape))
    print('>> Y_train shape = '+str(Y_train.shape))
    print('>> X_test shape = '+str(X_test.shape if len(X_test)!=0 else X_test))
    print('>> Y_test shape = '+str(Y_test.shape if len(Y_test)!=0 else Y_test))
    print(">> the number of training data = " + str(X_train.shape[1]))
    print(">> the number of test data = " + str(X_test.shape[1] if len(X_test)!=0 else X_test))
    layerdims = [X_train.shape[0]] + hpar['hidden_layer_dims'] + [Y_train.shape[0]]
    print('>> the number of units of layers = '+str(layerdims))
    print(">> the number of epochs = " + str(hpar['num_epochs']))
    print(">> learning rate = " + str(hpar['learning_rate']))
    print(">> activation function of hidden layers = " + hpar['activation_hidden_layers'])
    print(">> activation function of the last layer = " + hpar['activation_last_layer'])
    print(">> mini batch size = " + str(hpar['minibatch_size']))
    print(">> optimizer = " + hpar['optimizer'])
    print(">> input normalization = " + str(hpar['input_normalization']))
    print(">> global random seed = " + str(hpar['global_random_seed']))
    #------------------------------------------------------------------------------------
    ops.reset_default_graph()                         # to be able to rerun the model without overwriting tf variables
    tf.set_random_seed(hpar['global_random_seed'])                             # to keep consistent results
#    seed = 3                                          # to keep consistent results
    seed = hpar['global_random_seed']
    (n_x, m) = X_train.shape                          # (n_x: input size, m : number of examples in the train set)
    n_y = Y_train.shape[0]                            # n_y : output size
    costs = []                                        # To keep track of the cost
    
    X, Y = create_placeholders(n_x, n_y) # Create Placeholders of shape (n_x, n_y)
    parameters = initialize_parameters(layerdims, hpar['global_random_seed']) # Initialize parameters
    ZL = forward_propagation(X, parameters, hpar) # Forward propagation: Build the forward propagation in the tensorflow graph
    cost = compute_cost(ZL, Y, hpar) # Cost function: Add cost function to tensorflow graph
    if hpar['optimizer'] == 'gradient descent':
        optimizer = tf.train.GradientDescentOptimizer(learning_rate = hpar['learning_rate']).minimize(cost) # Backpropagation: Define the tensorflow optimizer. Use an AdamOptimizer.
    elif hpar['optimizer'] == 'momentum':
        optimizer = tf.train.MomentumOptimizer(learning_rate = hpar['learning_rate']).minimize(cost) # Backpropagation: Define the tensorflow optimizer. Use an AdamOptimizer.
    elif hpar['optimizer'] == 'adam':
        optimizer = tf.train.AdamOptimizer(learning_rate = hpar['learning_rate']).minimize(cost) # Backpropagation: Define the tensorflow optimizer. Use an AdamOptimizer.
    init = tf.global_variables_initializer() # Initialize all the variables
    
    with tf.Session() as sess: # Start the session to compute the tensorflow graph
        sess.run(init) # Run the initialization
        tic = time.time()
#        for epoch in range(0, hpar['num_epochs']): # Do the training loop        
        for epoch in range(1, hpar['num_epochs']+1): # Do the training loop
            epoch_cost = 0.                       # Defines a cost related to an epoch
            num_minibatches = max(1, int(m / hpar['minibatch_size'])) # number of minibatches of size minibatch_size in the train set
            seed = seed + 1
            minibatches = random_mini_batches(X_train, Y_train, hpar['minibatch_size'], seed)

            for minibatch in minibatches:
                (minibatch_X, minibatch_Y) = minibatch # Select a minibatch
                # IMPORTANT: The line that runs the graph on a minibatch.
                _ , minibatch_cost = sess.run([optimizer, cost], feed_dict={X:minibatch_X, Y:minibatch_Y}) # Run the session to execute the "optimizer" and the "cost", the feedict should contain a minibatch for (X,Y).
#                epoch_cost += minibatch_cost / num_minibatches
                epoch_cost += minibatch_cost * (minibatch_X.shape[1] / m)
            costs.append(epoch_cost)
            # Print the cost every epoch
#            if hpar['print_cost'] > 0 and epoch % hpar['print_cost'] == 0:
            if hpar['print_cost'] > 0 and ((epoch == 1) or (epoch % hpar['print_cost'] == 0) or (epoch == hpar['num_epochs'])):
                toc = time.time()
                print (">> epoch =  %i; cost = %f; time = %f sec" % (epoch, epoch_cost, toc-tic))
                tic = time.time()
        ## plot the cost
        #plt.plot(np.squeeze(costs))
        #plt.ylabel('cost')
        #plt.xlabel('iterations (per tens)')
        #plt.title("Learning rate =" + str(hpar['learning_rate']))
        #plt.show()
        # lets save the parameters in a variable
        parameters = sess.run(parameters)
        #print(">> Parameters have been trained!")
        # Calculate the correct predictions
        #correct_prediction = tf.equal(tf.argmax(ZL), tf.argmax(Y))
        ## Calculate accuracy on the test set
        #accuracy = tf.reduce_mean(tf.cast(correct_prediction, "float"))
        #print(">> Train Accuracy:", accuracy.eval({X: X_train, Y: Y_train}))
        #print(">> Test Accuracy:", accuracy.eval({X: X_test, Y: Y_test}))
    Yp_train, accuracy_train = predict(X_train, Y_train, parameters, hpar)
    Yp_test, accuracy_test = predict(X_test, Y_test, parameters, hpar)
    print(">> Train Accuracy:" + str(accuracy_train))
    print(">> Test Accuracy:" + str(accuracy_test))
    parameters['W1'], parameters['b1'] = denormalize_W1_b1(parameters['W1'], parameters['b1'], Xu, Xs) # denormalization
    print('> Deep Neural Network (tensorflow) ended ...')
    dout = {'parameters': parameters,
            'costs': costs,
            'Yp_train': Yp_train,
            'Yp_test': Yp_test}
    return dout
def predict(X_data, Y_data, parameters, hpar):
    para = {}
    for i in range(1, (len(parameters) // 2) + 1):
        para['W%d'%i] = tf.constant(parameters['W%d'%i], dtype = tf.float32)
        para['b%d'%i] = tf.constant(parameters['b%d'%i], dtype = tf.float32)
    xx = tf.constant(X_data, dtype = tf.float32)
    zl = forward_propagation(xx, para, hpar)
    sess = tf.Session()
    if hpar['activation_last_layer'] == 'sigmoid':
        al = tf.nn.sigmoid(zl)
        AL = sess.run(al)
        Yp = np.int64(AL > 0.5)
        YY = np.int64(Y_data > 0.5)
        accuracy = np.mean(np.all(YY == Yp, axis=0))
    elif hpar['activation_last_layer'] == 'softmax':
        ZL = sess.run(zl)
        YL = np.argmax(ZL, axis = 0)
        YY = np.argmax(Y_data, axis = 0)
        accuracy = np.mean(np.int64(YY == YL))
        Yp = convert_to_one_hot(YL, Y_data.shape[0])
    sess.close()
    return Yp, accuracy


# In[4]:


if False:
    M = 1000   # the number of samples
    np.random.seed(1)
    X = np.random.rand(2, M)
    Y = np.zeros((1, M))
    #Y[0, (X[1,:] - X[0,:] + 0.3) > 0 ] = 1  # line
    #Y[0, (X[0,:]**2 + X[1,:]**2 - 0.7**2) > 0 ] = 1  # semi-circle
    #Y[0, ((X[0,:]-0.5)**2 + (X[1,:]-0.5)**2 - 0.3**2) > 0] = 1  # circle
    #Y[0, (((X[0,:]-0.5)/0.4)**2 + ((X[1,:]-0.2)/0.05)**2 - 1) > 0] = 1  # ellipse
    Y[0, ((((X[0,:]-0.5)/0.4)**2 + ((X[1,:]-0.2)/0.05)**2 - 1) > 0) & (((X[0,:]-0.7)**2 + (X[1,:]-0.7)**2 - 0.1**2) > 0)] = 1  # ellipse & circle
    #-----------------------------------------------------------------
    rn = np.random.permutation(M)
    NofT = int(M * 0.7) # number of train samples
    X_train = X[:,rn[0:NofT]]
    Y_train = Y[:,rn[0:NofT]]
    X_test = X[:,rn[NofT:]]
    Y_test = Y[:,rn[NofT:]]
    #-----------------------------------------------------------------
    hpar = {}
    hpar['hidden_layer_dims'] = [10,5,2]
    hpar['activation_hidden_layers'] = ['sigmoid', 'tanh', 'relu'][1]
    hpar['activation_last_layer'] = ['sigmoid', 'softmax'][0]
    hpar['learning_rate'] = 0.1
    hpar['num_epochs'] = 10000
    hpar['minibatch_size'] = 999999
    hpar['optimizer'] = ["gradient descent", "momentum", "adam"][0]
    hpar['print_cost'] = 1000 # print epoch step; no print when it is zero
    hpar['input_normalization'] = True
    hpar['global_random_seed'] = 1
    dout = train(X_train, Y_train, X_test, Y_test, hpar)
    #-----------------------------------------------------------------
    Yp_train, accuracy_train = predict(X_train, Y_train, dout['parameters'], hpar)
    Yp_test, accuracy_test = predict(X_test, Y_test, dout['parameters'], hpar)
    print("- Train Accuracy:" + str(accuracy_train))
    print("- Test Accuracy:" + str(accuracy_test))  


# In[5]:


if False:
    M = 1000   # the number of samples
    np.random.seed(1)
    X = np.random.rand(2, M)
    Y = np.zeros((1, M))
    #Y[0, (X[1,:] - X[0,:] + 0.3) > 0 ] = 1  # line
    #Y[0, (X[0,:]**2 + X[1,:]**2 - 0.7**2) > 0 ] = 1  # semi-circle
    #Y[0, ((X[0,:]-0.5)**2 + (X[1,:]-0.5)**2 - 0.3**2) > 0] = 1  # circle
    #Y[0, (((X[0,:]-0.5)/0.4)**2 + ((X[1,:]-0.2)/0.05)**2 - 1) > 0] = 1  # ellipse
    Y[0, ((((X[0,:]-0.5)/0.4)**2 + ((X[1,:]-0.2)/0.05)**2 - 1) > 0) & (((X[0,:]-0.7)**2 + (X[1,:]-0.7)**2 - 0.1**2) > 0)] = 1  # ellipse & circle
    #-----------------------------------------------------------------
    rn = np.random.permutation(M)
    NofT = int(M * 0.7) # number of train samples
    X_train = X[:,rn[0:NofT]]
    Y_train = Y[:,rn[0:NofT]]
    X_test = X[:,rn[NofT:]]
    Y_test = Y[:,rn[NofT:]]
    #-----------------------------------------------------------------
    hpar = {}
    hpar['hidden_layer_dims'] = [10,5,2]
    hpar['activation_hidden_layers'] = ['sigmoid', 'tanh', 'relu'][2]
    hpar['activation_last_layer'] = ['sigmoid', 'softmax'][0]
    hpar['learning_rate'] = 0.001
    hpar['num_epochs'] = 2000
    hpar['minibatch_size'] = 64
    hpar['optimizer'] = ["gradient descent", "momentum", "adam"][2]
    hpar['print_cost'] = 1000 # print epoch step; no print when it is zero
    hpar['input_normalization'] = True
    hpar['global_random_seed'] = 1
    dout = train(X_train, Y_train, X_test, Y_test, hpar)
    #-----------------------------------------------------------------
    Yp_train, accuracy_train = predict(X_train, Y_train, dout['parameters'], hpar)
    Yp_test, accuracy_test = predict(X_test, Y_test, dout['parameters'], hpar)
    print("- Train Accuracy:" + str(accuracy_train))
    print("- Test Accuracy:" + str(accuracy_test))      


# In[6]:


if False:
    dataset_file_path = './dataset/coursera_SIGNS_dataset/dataset_(X_train, Y_train, X_test, Y_test).pckl'
    f = open(dataset_file_path, 'rb')
    (X_train_orig, Y_train_orig, X_test_orig, Y_test_orig) = pickle.load(f)
    f.close()
    classes = np.array([0,1,2,3,4,5])
    #-------------------------------------------------------
    # Example of a picture
    #index = 0
    #plt.imshow(X_train_orig[index])
    #print ("y = " + str(np.squeeze(Y_train_orig[:, index])))
    #-------------------------------------------------------
    # Flatten the training and test images
    X_train_flatten = X_train_orig.reshape(X_train_orig.shape[0], -1).T
    X_test_flatten = X_test_orig.reshape(X_test_orig.shape[0], -1).T
    # Normalize image vectors
    X_train = X_train_flatten/255.
    X_test = X_test_flatten/255.
    # Convert training and test labels to one hot matrices
    Y_train = convert_to_one_hot(Y_train_orig, 6)
    Y_test = convert_to_one_hot(Y_test_orig, 6)
    #-----------------------------------------------------------------------------------
    hpar = {}
    hpar['hidden_layer_dims'] = [25,12]
    hpar['activation_hidden_layers'] = ['sigmoid', 'tanh', 'relu'][2]
    hpar['activation_last_layer'] = ['sigmoid', 'softmax'][1]
    hpar['learning_rate'] = 0.0001
    hpar['num_epochs'] = 1500
    hpar['minibatch_size'] = 32
    hpar['optimizer'] = ["gradient descent", "momentum", "adam"][2]
    hpar['print_cost'] = 100 # print epoch step; no print when it is zero
    hpar['input_normalization'] = False
    hpar['global_random_seed'] = 1
    dout = train(X_train, Y_train, X_test, Y_test, hpar)
    #-----------------------------------------------------------------
    Yp_train, accuracy_train = predict(X_train, Y_train, dout['parameters'], hpar)
    Yp_test, accuracy_test = predict(X_test, Y_test, dout['parameters'], hpar)
    print("- Train Accuracy:" + str(accuracy_train))
    print("- Test Accuracy:" + str(accuracy_test))      


# In[ ]:







# https://medium.com/@joel.barmettler/how-to-upload-your-python-package-to-pypi-65edc5fe9c56

from setuptools import setup

setup(name='bradk',
      version='0.11',
      description='The bradk packages developed by Brad Kim',
      url='http://github.com/bomsoo-kim/packages',
      author='Brad Kim',
      author_email='bk2717@columbia.edu',
      license='MIT',
      packages=['bradk', 'bradk.model_selection', 'bradk.finance', 'bradk.neural_networks'],
      zip_safe=False)



# coding: utf-8

# In[1]:


import numpy as np
import matplotlib.pyplot as plt
import time
import math
#-------------------------------
import keras
from keras.models import Sequential
from keras.layers import Dense, Activation, BatchNormalization, Conv1D, MaxPooling1D, Flatten
from keras.optimizers import SGD, Adam


# In[2]:


def train(x_train, y_train, x_test, y_test, hpar):
    input_shape = x_train.shape[1:]
    output_num = y_train.shape[1]
    if hpar['activation_last_layer'] == 'sigmoid': # loss function
        loss = 'binary_crossentropy'
    elif hpar['activation_last_layer'] == 'softmax':
        loss = 'categorical_crossentropy'
    if hpar['optimizer'] == 'gradient descent': # optimizer
        optimizer = SGD(lr=hpar['learning_rate'])
    elif hpar['optimizer'] == 'adam':
        optimizer = Adam(lr=hpar['learning_rate'])
    #-----------------------------------------------------
    model = Sequential()
    model.add(BatchNormalization(input_shape=input_shape))
    for i in range(0, len(hpar['layer_conv_filter'])):
        conv_ksize = hpar['layer_conv_filter'][i][0]
        conv_filters = hpar['layer_conv_filter'][i][1]
        conv_stride = hpar['layer_conv_stride'][i]
        conv_pad = hpar['layer_conv_padding'][i]
        pool_ksize = hpar['layer_pool_filter'][i]
        pool_stride = hpar['layer_pool_stride'][i]
        pool_pad = hpar['layer_pool_padding'][i]
        model.add(Conv1D(conv_filters, conv_ksize, strides=conv_stride, padding=conv_pad))
        model.add(BatchNormalization())
        model.add(Activation(hpar['activation_hidden_layers'])) # 'sigmoid', 'tanh', 'relu'
        model.add(MaxPooling1D(pool_size=pool_ksize, strides=pool_stride, padding=pool_pad))
    model.add(Flatten())
    for i in range(len(hpar['hlayer_fully_connected'])):
        model.add(Dense(hpar['hlayer_fully_connected'][i]))
        model.add(BatchNormalization())
        model.add(Activation(hpar['activation_hidden_layers'])) # 'sigmoid', 'tanh', 'relu'
    model.add(Dense(output_num))
    model.add(BatchNormalization())
    model.add(Activation(hpar['activation_last_layer']))
    model.compile(optimizer = optimizer, loss = loss, metrics = ['accuracy'])
    #-----------------------------------------------------
    model.summary()
    loss = []
    accu = []
    for i in range(1, 1 + max(1, hpar['num_epochs'] // hpar['print_cost'])):
        tic = time.time()
        fout = model.fit(x_train, y_train, epochs=hpar['print_cost'], batch_size=hpar['minibatch_size'], verbose=0)
        toc = time.time()
        loss = loss + fout.history['loss']
        accu = accu + fout.history['acc']
        score_train = model.evaluate(x_train, y_train, batch_size=hpar['minibatch_size'], verbose=0)
        score_test = model.evaluate(x_test, y_test, batch_size=hpar['minibatch_size'], verbose=0)
        print('>> epochs = ', (i*hpar['print_cost']),'; Train loss/acc = %f/%f'%(score_train[0],score_train[1]), '; Test loss/acc = %f/%f'%(score_test[0],score_test[1]), '; time = %f sec '%(toc-tic))
    mout = {'model': model,
            'loss': loss,
            'accu': accu}
    return mout


# In[3]:


if False:
    hpar = {}
    hpar['layer_conv_filter'] = [[7,8], [3,16]] # [height, width, channel]
    hpar['layer_conv_stride'] = [1, 1] # [height, width]
    hpar['layer_conv_padding'] = ['SAME', 'SAME']
    hpar['layer_pool_filter'] = [4, 4] # [height, width]
    hpar['layer_pool_stride'] = [4, 4] # [height, width]
    hpar['layer_pool_padding'] = ['VALID', 'VALID']
    hpar['hlayer_fully_connected'] = []
    hpar['activation_hidden_layers'] = ['sigmoid', 'tanh', 'relu'][1]
    hpar['activation_last_layer'] = ['sigmoid', 'softmax'][1]
    hpar['learning_rate'] = 0.4
    hpar['num_epochs'] = 3000
    hpar['minibatch_size'] = 999999
    hpar['optimizer'] = ["gradient descent", "momentum", "adam"][0]
    hpar['print_cost'] = 100 # print epoch step; no print when it is zero
    mout = train(x_train, y_train_one_hot, x_test, y_test_one_hot, hpar)
    #-------------------------------------------------------------
    plt.plot(mout['loss'])
    plt.show()
    plt.plot(mout['accu'])
    plt.show()
    #-------------------------------------------------------------
    yp_test_one_hot = mout['model'].predict(x_test)
    yp_test = np.argmax(yp_test_one_hot, axis = 1).reshape(y_test.shape)
    accuracy_test = np.mean(y_test == yp_test)
    print(accuracy_test)  


# In[ ]:








# coding: utf-8

# In[1]:


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates


# In[2]:


def get_RSI(xx, AVGPERIODS): # Relative Strength Index (RSI)
    # ref) http://stockcharts.com/school/doku.php?id=chart_school:technical_indicators:relative_strength_index_rsi
    assert(len(xx.shape) == 1) # assert that xx be rank 0
    Change = np.full(xx.shape, 0) # initialize
    Change[1:] = np.diff(xx, n=1)
    Gain = np.full(xx.shape, 0) # initialize
    Gain[Change > 0] = Change[Change > 0]
    Loss = np.full(xx.shape, 0) # initialize
    Loss[Change < 0] = np.absolute(Change[Change < 0])
    AvgGain = np.full(xx.shape, np.nan) # initialize
    AvgLoss = np.full(xx.shape, np.nan) # initialize
    RS = np.full(xx.shape, np.nan) # ref) https://stackoverflow.com/questions/1704823/initializing-numpy-matrix-to-something-other-than-zero-or-one
    RSI = np.full(xx.shape, np.nan)
    #----------------------------------------------------------------
    AvgGain[AVGPERIODS-1] = np.mean(Gain[:AVGPERIODS])
    AvgLoss[AVGPERIODS-1] = np.mean(Loss[:AVGPERIODS])
    RS[AVGPERIODS-1] = AvgGain[AVGPERIODS-1] / AvgLoss[AVGPERIODS-1]
    RSI[AVGPERIODS-1] = 100 - 100 / (1 + RS[AVGPERIODS-1])
    for n in range(AVGPERIODS, len(xx)):
        AvgGain[n] = (AvgGain[n-1]*(AVGPERIODS - 1) + Gain[n]) / AVGPERIODS
        AvgLoss[n] = (AvgLoss[n-1]*(AVGPERIODS - 1) + Loss[n]) / AVGPERIODS
        RS[n] = AvgGain[n] / AvgLoss[n]
        RSI[n] = 100 - 100 / (1 + RS[n])    
    return RSI
def get_EMA(xx, AVGPERIODS): # Moving Averages - Simple and Exponential
    # ref) http://stockcharts.com/school/doku.php?id=chart_school:technical_indicators:moving_averages
    assert(len(xx.shape) == 1) # assert that xx be rank 0
    Multiplier = 2 / (AVGPERIODS + 1)
    EMA = np.full(xx.shape, np.nan)
    if AVGPERIODS < len(xx):
        EMA[AVGPERIODS-1] = np.mean(xx[:AVGPERIODS])
    for n in range(AVGPERIODS, len(xx)):
        EMA[n] = (xx[n] - EMA[n-1])*Multiplier + EMA[n-1]
    return EMA
def get_MACD(xx, AVGPERIODS1, AVGPERIODS2, AVGPERIODS3): # MACD (Moving Average Convergence/Divergence Oscillator)
    # ref) http://stockcharts.com/school/doku.php?st=macd&id=chart_school:technical_indicators:moving_average_convergence_divergence_macd
    assert(len(xx.shape) == 1) # assert that xx be rank 0
    MACD = get_EMA(xx, AVGPERIODS1) - get_EMA(xx, AVGPERIODS2)
    SIGNAL = np.full(xx.shape, np.nan); # initialize
    SIGNAL[~np.isnan(MACD)] = get_EMA(MACD[~np.isnan(MACD)], AVGPERIODS3);
    MACD_HIST = MACD - SIGNAL
    return MACD, SIGNAL, MACD_HIST
def get_OBV(price, volume): # On Balance Volume (OBV)
    # ref) https://stockcharts.com/school/doku.php?id=chart_school:technical_indicators:on_balance_volume_obv
    assert(len(price.shape) == 1) # make sure that the rank is 0
    assert(len(volume.shape) == 1) # make sure that the rank is 0
    assert(len(price.shape) == len(volume.shape)) # make sure that the ranks are the same
    up_down = np.diff(price, n=1) # detect the up and down of the price
    up_down[up_down > 0] = +1 # plus sign
    up_down[up_down < 0] = -1 # minus sign
    sign_volume = up_down * volume[1:] # exclude the first element of volume
    OBV = np.full(price.shape, np.nan) # initialize
    OBV[1:] = np.cumsum(sign_volume)
    return OBV
def get_MovingMomentumStrategy(prices, EMA_short_period = 15, EMA_long_period = 200, RSI_period = 14
                               , MACD_period1 = 12, MACD_period2 = 26, MACD_period3 = 9):
    EMA_short = get_EMA(prices, EMA_short_period) # 15
    EMA_long = get_EMA(prices, EMA_long_period) # 200
    RSI = get_RSI(prices, RSI_period) # 14
    MACD, SIGNAL, MACD_HIST = get_MACD(prices, MACD_period1, MACD_period2, MACD_period3) # 12, 26, 9
    return EMA_short, EMA_long, RSI, MACD_HIST
def plot_MovingMomentumStrategy(dates, prices, EMA_short, EMA_long, RSI, MACD_HIST):
    # ref) https://matplotlib.org/examples/pylab_examples/subplots_demo.html
    # ref) https://matplotlib.org/faq/usage_faq.html
    f, (ax1, ax2, ax3) = plt.subplots(3, sharex=True, sharey=True, figsize=(16, 15))
    ax1.plot_date(dates, prices, 'k.-') # prices vs. dates
    ax1.plot_date(dates, EMA_short, 'r-') # prices vs. short MV
    ax1.plot_date(dates, EMA_long, 'b-') # prices vs. long MV
    ax1.set_ylabel('Price')
    ax1.xaxis.grid(which='major')
    ax1.xaxis.grid(which='minor')
    ax1.yaxis.grid()
    ax2.plot_date(dates, prices, 'k.-') # prices vs. dates
    ax2.set_ylabel('Price')
    ax22 = ax2.twinx()
    ax22.plot_date(dates, RSI, 'r-') # RSI vs. dates
    ax22.plot_date(dates, 70 + np.zeros(dates.shape), 'b-.') # auxiliary line
    ax22.plot_date(dates, 50 + np.zeros(dates.shape), 'b-.') # auxiliary line
    ax22.plot_date(dates, 30 + np.zeros(dates.shape), 'b-.') # auxiliary line
    ax22.set_ylabel('RSI [%]')
    ax2.xaxis.grid(which='major')
    ax2.xaxis.grid(which='minor')
    ax2.yaxis.grid()
    ax3.plot_date(dates, prices, 'k.-') # prices vs. dates
    ax3.set_ylabel('Price')
    ax32 = ax3.twinx()
#    ax32.bar(dates, MACD_HIST, 1.0, color='c') # MACD_HIST vs. dates       # TOO SLOW
    ax32.plot_date(dates, MACD_HIST, 'b-') # MACD_HIST vs. dates
    ax32.plot_date(dates, 0 + np.zeros(dates.shape), 'b-') # auxiliary line
    ax32.set_ylabel('MACD_HIST')
    ax3.xaxis.grid(which='major')
    ax3.xaxis.grid(which='minor')
    ax3.yaxis.grid()
    ax3.xaxis.set_major_locator(mdates.YearLocator())
#    ax3.xaxis.set_major_formatter(mdates.DateFormatter('%Y '))
    ax3.xaxis.set_minor_locator(mdates.MonthLocator())
#    ax3.xaxis.set_minor_formatter(mdates.DateFormatter('%m'))
    f.autofmt_xdate()
    f.subplots_adjust(hspace=0)
    #plt.setp([a.get_xticklabels() for a in f.axes[:-1]], visible=False)
#    plt.xlim(min_date, max_date)
    plt.show()





# coding: utf-8

# In[1]:


import pandas as pd
pd.core.common.is_list_like = pd.api.types.is_list_like # https://stackoverflow.com/questions/50394873/import-pandas-datareader-gives-importerror-cannot-import-name-is-list-like
import pandas_datareader.data as web


# In[2]:


# http://pandas-datareader.readthedocs.io/en/latest/remote_data.html
# https://ntguardian.wordpress.com/2016/09/19/introduction-stock-market-data-python-1/
def download_tables(ticker, start=pd.Timestamp('1800-01-01'), end=pd.Timestamp.today()): # read table data from the star page to end page
    df = web.DataReader(ticker, 'fred', start, end)
    return df


# In[3]:


def update_data_in_db(filepath, ticker, COL_DATE = 'DATE'):
    try:
        df = pd.read_csv(filepath) # read the already sorted data
        df[COL_DATE] = pd.to_datetime(df[COL_DATE])
        latestDate = df.iloc[-1][COL_DATE] # temporarily save the latest date
        if latestDate < pd.Timestamp.today():
            start=latestDate + pd.Timedelta(seconds=1)
            end=pd.Timestamp.today()
            df0 = download_tables(ticker, start, end)
            df0.reset_index(inplace=True)
            df = pd.concat([df, df0])
    except:
        print("File not found in the database. Now, searching the internet to download information...")
        df = download_tables(ticker) # download the data from the website
        df.reset_index(inplace=True)
        print("Downloading completed.")
    df.sort_values(COL_DATE, inplace = True) # sort in order of date & time
    df.to_csv(filepath, index=False) # save the data to db
    return df


# In[4]:


if False:
    ticker = 'UNRATE'
    df = download_tables(ticker)

    import matplotlib.pyplot as plt
    get_ipython().run_line_magic('matplotlib', 'inline')
    df['UNRATE'].plot(grid = True)
if False:
    df = update_data_in_db('c:/Users/bomso/bomsoo1/project/python/unemply.csv', 'UNRATE')
    print(df.iloc[-35:])

